"""
MHD Radiative Transfer Loader Module
====================================

This module provides functionality for computing single-fluid magnetohydrodynamics (MHD)
quantities related to radiative transfer synthesis and spectral line analysis.

Main Components
--------------
MhdRadiativeLoader : class
    Primary class for radiative transfer calculations in MHD simulations, including:
    - Emission measure (EM) with optional TRAC correction and opacity
    - Velocity-Differential Emission Measure (VDEM)
    - Spectral line synthesis and Doppler velocity analysis
    - G(T) contribution functions from ChiantiPy

Helper Functions
---------------
lambda_to_doppler : Convert wavelength to Doppler velocity units
doppler_to_lambda : Convert Doppler velocity to wavelength units
read_qloss : Read radiative loss tables from file
interp_masked : Array masking with interpolation for VDEM binning

External Dependencies
--------------------
- muse.instr.utils : MUSE instrument utilities for spectral synthesis
- ChiantiPy.core : Atomic database for spectral line calculations

Note: Some external dependencies may be optional and will fail gracefully
with ImportFailed if not available.
"""

# builtins
import os
import shutil
import datetime as datetime

# public/common external (and, part of PlasmaCalcs requirements already)
import xarray as xr
import numpy as np

# private/external (not part of PlasmaCalcs requirements)
from ..tools import ImportFailed  # <-- helps if import fails
try:
    from muse.instr import utils as muse_instr_utils  
except ImportError as err:
    muse_instr_utils = ImportFailed("muse.instr.utils", err=err, locals=locals(), abbrv='muse_instr_utils')

try:
    import ChiantiPy.core as ch
except ImportError as err:
    ch = ImportFailed("ChiantiPy.core", err=err, locals=locals(), abbrv='ch')

# internal
from .mhd_bases import MhdBasesLoader
from ..defaults import DEFAULTS
from ..tools import (
    alias, simple_property, format_docstring,
    pcAccessor,
    xr1d, _xarray_save_prep,
    xarray_result_size_check,
    xarray_multiply_then_sum_along,
    xarray_moments,
    xarray_update_call_history,
)
from ..errors import (
    InputError, InputConflictError, InputMissingError,
    ComponentValueError, DimensionalityError,
)


### --------------------- Helper functions --------------------- ###

def lambda_to_doppler(wave, wave0):
    """
    Convert wavelength to Doppler velocity units (km/s).

    Parameters
    ----------
    wave : array-like
        Observed wavelength(s) [same units as wave0, e.g., Angstroms].
    wave0 : float
        Rest wavelength.

    Returns
    -------
    array-like
        Doppler velocity in km/s.
    """
    cc = 2.99792458e10 / 1e5
    return (wave / wave0 - 1.0) * cc  # outputs in km/s


def doppler_to_lambda(doppler, wave0):
    """
    Convert wavelength to Doppler velocity units (km/s).

    Parameters
    ----------
    doppler : array-like
        Doppler velocity in km/s.
    wave0 : float
        Rest wavelength.

    Returns
    -------
    array-like
        Observed wavelength(s) [same units as wave0, e.g., Angstroms]
    """
    cc = 2.99792458e10 / 1e5
    return wave0 * (1.0 + doppler / cc)  # outputs in units of wave0


def read_qloss(file="Radloss_Chianti.dat"):
    """
    Read radiative loss table from file.

    Parameters
    ----------
    file : str
        Filename for the radiative loss table.

    Returns
    -------
    tg : ndarray
        Temperature grid.
    qloss : ndarray
        Radiative loss values.
    """
    f = open("Radloss_Chianti.dat", "r")
    first_line = f.readline()
    ntg = int(first_line.strip())
    tg = np.zeros(ntg)
    qloss = np.zeros(ntg)

    for ii, line in enumerate(f):
        tg[ii], qloss[ii] = np.float64(line.strip().split())

    return tg, qloss

@pcAccessor.register('interp_masked', totype='array')   # <-- would be fine, if desired!
def interp_masked(array, bins, dim, *, mode='correct', return_weights=False):
    """
    Array masked to bins, possibly with interpolation correction and normalization factors,
        for use in (Velocity) Differential Emission Measure ((V)DEM) binning.
    The result has `array` dims plus `bins` dim, with values from 0 to 1.

    Conceptually - basically, this just tells whether `array` overlaps with each bin.
        E.g. result.isel(x=i, y=j, bindim=k) tells whether the (i,j)th x,y point overlaps with bin k.
    Actually - it may include an interpolation correction factor,
        accounting for adjacent cell values, and any jumps larger than bin widths.
        E.g. result.isel(x=i, bindim=k) tells how much of bin k is covered from the ith to (i+1)th x.
    Additionally - if return_weights, this method also returns xn, which can be applied to a value,
        to properly account for interpolation between adjacent cells.

    See Rempel et al. (2017) for details.
    (Note - Rempel+2017 formulas are slightly different than those here,
        because those formulas solve a slightly different problem.
        Testing those formulas here led to more numerical artifacts.)

    If binning multiple quantities (e.g., temperature and velocity), the canonical pattern to use is:
        mask1, xn = self.interp_masked(quantity1, bins1, return_weights='exp')
        mask2 = self.interp_masked(quantity2, bins2)
        mask3 = self.interp_masked(quantity3, bins3)
        ...   # etc. i.e., only the first call computes interpolation weights.
        emiss_interped = emiss * (1.0 - xn) + emiss.roll({x:1}) * xn

    Note - the result's first value along dim may be unreliable (if non-periodic array),
        as it relies on array.roll({dim:1}).

    Parameters
    ----------
    array : xarray.DataArray
        variable to be interp-masked. E.g., array of log10(temperature) values.
    bins : xarray.DataArray
        1D array of (evenly-spaced) bin centers (internally, assumes bin width = (bins[1]-bins[0])/2).
    dim : str
        apply the correction along this dimension of array.
        For VDEMS, this is the dimension along which the LOS integration will be performed.
    mode: 'correct' or 'simple'
        tells how to compute the result.
        'correct' --> includes correction factor.
        'simple' --> exclude correction factor, just tells whether each point falls within each bin.
                    Equivalent to: result = (bins - delta <= array) & (array < bins + delta),
                        where delta = (bins[1] - bins[0]) / 2 == bin width.
                    This can be much faster than mode='correct', but produces noisier results.
                    (Also, only allowed if return_weights=False).
    return_weights : False, True, or 'exp'
        whether to also return interpolation weights xn for each bin.
        False --> returns masked
        True --> returns (masked, xn)
        'exp' --> returns (masked, xn) but compute xn using (10**value, 10**bins) not (value, bins).

    Returns
    -------
    masked : xarray.DataArray
        array of fn * overlaps, where overlaps = whether each cell overlaps with each bin,
        and fn = normalization factor for each bin, clipped to [0, 1].
        (Or, if mode=='simple', this is just (bins - delta <= array) & (array < bins + delta).)
    xn : xarray.DataArray
        Interpolation weight for each bin, with values clipped to [0, 1].
        (Only returned if return_weights.)
    """
    ### checks ###
    if not isinstance(bins, xr.DataArray):
        raise TypeError("interp_masked: bins must be an xarray.DataArray")
    if bins.ndim != 1:
        raise DimensionalityError(f'interp_masked: expected 1D bins; got bins.ndim={bins.ndim}')
    if isinstance(return_weights, str) and return_weights != 'exp':
        raise InputError(f'return_weights should be False, True, or "exp"; got {return_weights!r}')
    if mode not in ('correct', 'simple'):
        raise InputError(f'mode should be "correct" or "simple"; got {mode!r}')
    if mode == 'simple' and return_weights:
        errmsg = f'mode="simple" only allowed if return_weights=False; got return_weights={return_weights!r}.'
        raise InputConflictError(errmsg)

    ### initial setup ###
    xarray_result_size_check(array, bins)  # if result would be too huge, make crash instead!
    delta = (bins.values[1] - bins.values[0]) / 2
    x = dim
    value = array

    ### mode=='simple' case ###
    if mode == 'simple':
        inbins = (value >= bins - delta) & (value < bins + delta)
        return inbins

    ### mode=='correct' ###
    value1 = value.roll({x:1})
    minvalue = np.minimum(value, value1)
    maxvalue = np.maximum(value, value1)
    # overlaps = "whether any part of the segment (from minvalue to maxvalue) overlaps the bin."
    overlaps = (maxvalue >= bins - delta) & (minvalue < bins + delta)
    clipmin = minvalue.clip(min=bins - delta)
    clipmax = maxvalue.clip(max=bins + delta)
    fn = np.abs((clipmax - clipmin) / (value1 - value))  # correction coefficient
    fn = fn.fillna(0).clip(0, 1)
    masked = overlaps * fn

    if return_weights:
        ### interpolation weights ###
        if return_weights=='exp':
            xn = (10**value - 10**bins) / np.abs(10**value - 10**value1)
        else:
            xn = (value - bins) / np.abs(value - value1)
        xn = xn.fillna(0).clip(0, 1)
        return masked, xn
    else:  # no need to compute weights
        return masked


### --------------------- MhdRadiativeLoader --------------------- ###

class MhdRadiativeLoader(MhdBasesLoader):
    """
    Loader for computing single-fluid MHD quantities related to radiative transfer.

    This class extends MhdBasesLoader to provide specialized functionality for
    radiative transfer calculations in magnetohydrodynamics (MHD) simulations,
    including:

    - **Emission Measure (EM)**: Calculations with optional TRAC (Transition Region
      Adaptive Conduction) correction and opacity effects
    - **Velocity-Differential Emission Measure (VDEM)**: Temperature and velocity
      binning for spectral line analysis
    - **Spectral Line Synthesis**: Full and partial spectra generation with instrument
      profiles and multiple emission line support
    - **Doppler Shifts**: Line-of-sight velocity analysis via Doppler shift calculations
    - **Intensity Moments**: Computation of spectral line moments (intensity, mean
      velocity, width)

    Key Features
    ------------
    - Configurable emission measure modes (with/without TRAC, with/without opacity)
    - Flexible temperature and velocity binning for VDEM calculations
    - Multiple VDEM computation modes: interpolated, non-interpolated, and fully interpolated
    - Support for photosphere masking in calculations
    - Integration with MUSE instrument utilities and ChiantiPy for spectral synthesis
    - Efficient handling of large datasets through xarray integration

    Attributes
    ----------
    emiss_mode : str
        Emission measure calculation mode ('notrac_noopa', 'trac_noopa', 
        'notrac_opa', 'trac_opa')
    emiss_trac : bool
        Whether to include TRAC correction in emission measure calculations
    emiss_opa : bool
        Whether to include opacity effects in emission measure calculations
    emiss_norm : float
        Emission measure normalization factor (default: 1e27)
    vdem_mode : str
        VDEM calculation mode ('interp', 'nointerp', 'allinterp')
    rcoords_logT : array-like
        Log10(temperature [K]) values for binning
    rcoords_vdop_kms : array-like
        Velocity/Doppler shift values in km/s for binning
    vdem_logT_min : float
        Minimum log10(temperature [K]) threshold for VDEM calculations

    See Also
    --------
    MhdBasesLoader : Parent class providing base MHD functionality
    """

    # # # HELPER METHODS # # #
    def _radiative_assert_single_component(self):
        '''asserts self.current_n_component() == 1; raise ComponentValueError otherwise.
        The error message will be particularly helpful and give options for how to fix the error.
        '''
        # [TODO] probably shouldn't exist anymore; relevant methods should use _los and/or _loi patterns.
        if self.current_n_component() != 1:
            errmsg = ("Inside of a MhdRadiativeLoader method, requiring a single component, "
                      f"but got self.component={self.component}.\nConsider specifying component, "
                      "e.g. via self.component='x', or self(f'{var}_y'), or self(var, component='z')")
            raise ComponentValueError(errmsg)

    # # # VARIOUS SETTINGS # # #
    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['emiss_mode', 'vdem_logT', 'vdem_loopdim'] + any values from super().
        '''
        return ['emiss_mode', 'vdem_logT', 'vdem_loopdim'] + super()._extra_kw_for_quantity_loader_call

    # # # EMISSION MEASURE # # #

    cls_behavior_attrs.register('emiss_trac', default=False)
    emiss_trac = simple_property('_emiss_trac', default=False,
        doc='''whether to include TRAC correction in emission measure (self('emiss')).''')

    cls_behavior_attrs.register('emiss_opa', default=False)
    emiss_opa = simple_property('_emiss_opa', default=False,
        doc='''whether to include opacity in emission measure (self('emiss')).''')

    @property
    def emiss_mode(self):
        '''string telling emission measure mode.
        Depends on (and can be used to set) emiss_trac and emiss_opa.

        'notrac_noopa' <--> emiss_trac=False, emiss_opa=False
        'trac_noopa' <--> emiss_trac=True, emiss_opa=False
        'notrac_opa' <--> emiss_trac=False, emiss_opa=True
        'trac_opa' <--> emiss_trac=True, emiss_opa=True

        Setting string sets emiss_trac and emiss_opa accordingly.
        Getting string returns result according to emiss_trac and emiss_opa.

        (Currently included for backwards compatibility, might be removed eventually.)
        '''
        trac = self.emiss_trac
        opa = self.emiss_opa
        trac_str = 'trac' if trac else 'notrac'
        opa_str = 'opa' if opa else 'noopa'
        return f'{trac_str}_{opa_str}'
    @emiss_mode.setter
    def emiss_mode(self, value):
        EMISS_MODE_OPTIONS = ('notrac_noopa', 'trac_noopa', 'notrac_opa', 'trac_opa')
        if value not in EMISS_MODE_OPTIONS:
            raise InputError(f"Invalid emiss_mode {value!r}; expected one of {list(EMISS_MODE_OPTIONS.keys())}.")
        trac_str, opa_str = value.split('_')
        self.emiss_trac = (trac_str == 'trac')
        self.emiss_opa = (opa_str == 'opa')


    @known_var(attr_deps=[('emiss_trac', {True: ['emiss_trac'], False: ['emiss_notrac']}),
                          ('emiss_opa', {True: ['tau'], False: []})])
    def get_emiss(self):
        """Emission measure (EM) based on the selected emission mode (self.emiss_mode).
        Depends on self.emiss_mode; see self.EMISS_MODE_OPTIONS for details.

        Result is always in cgs, regardless of self.units.
        """
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):
            if self.emiss_trac:
                emiss = self('emiss_trac')
            else:
                emiss = self('emiss_notrac')
            if self.emiss_opa:
                tau = self('tau')
                # [TODO] emiss = self.loi_interp(emiss)  in this case...
                xarray_result_size_check(emiss, tau)  # prevent wayyyy too big result
                result = emiss * np.exp(-tau)
            else:
                result = emiss
            return result

    @known_var(deps=['ne', 'r'])
    def get_emiss_notrac(self):
        '''Emission measure in cgs without TRAC correction.
        Result is always in [cgs], regardless of self.units.

        nh = r / (n_per_nh * mh)
        emiss_notrac = ne * nh / self.emiss_norm [1/cm^6]
        '''
        self.assert_single_fluid_mode('emiss_notrac')
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):
            r_per_nH_tot = (self.elements.n_per_nH() * self.elements.m * self.u('amu')).sum()
            val = self('ne') * self('r') / (r_per_nH_tot * self.emiss_norm)
            return val

    cls_behavior_attrs.register('emiss_norm', default=1e27)
    emiss_norm = simple_property('_emiss_norm', default=1e27,
        doc='''Emission Measure normalization factor. Used in self('emiss')
        and related vars. emiss = unnormed_emiss / emiss_norm.''')

    cls_behavior_attrs.register('emiss_trac_coeff')
    emiss_trac_coeff = simple_property('_emiss_trac_coeff', default=3,
        doc='''TRAC coefficient for emiss_trac calculation.
         transition region using an adaptive conduction coefficient (Johnston et al. 2020).''')

    @known_var(deps=['emiss_notrac', 'B', 'T'], ignores_dims=['component'])
    def get_emiss_trac(self):
        '''Emission measure in cgs with TRAC correction.
        Result is always in [cgs], regardless of self.units.

        nh = r / (n_per_nh * mh)
        emiss_trac = ne * nh * trac_correction / normalization

        self.emiss_trac_coeff tells the TRAC coefficient (default 3) 
        '''
        self.assert_single_fluid_mode('emiss_trac')
        if not set(('x', 'y', 'z')) <= set(self.maindims):  # i.e., x, y, and z are not all in maindims:
            errmsg = f'emiss_trac requires 3D data (x, y, and z), but got self.maindims={self.maindims}.'
            raise DimensionalityError(errmsg)
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):
            emis = self('emiss_notrac')
            bx = self("B", component='x', units = 'raw')  # only using relative values of Bx,By,Bz so 'raw' is cheaper.
            by = self("B", component='y', units = 'raw')  #    also, 'cgs' B would complain, by default.
            bz = self("B", component='z', units = 'raw')
            tg = self('T')
            dx = self('maindims_x_dslice')  # dx accounting for self.slices appropriately for integral.
            dy = self('maindims_y_dslice')  # dy ^  (e.g. if slice(None,None,4), want ~4 times larger dy)
            dz = self('maindims_z_dslice')  # dz ^^

            tg_dep, qloss_dep = read_qloss()
            Lt = xr.DataArray(qloss_dep, coords={"T": tg_dep}, dims="T").interp(T=tg)

            Q_Chianti = emis * (Lt * self.emiss_norm)
            modb = np.sqrt(bx**2 + by**2 + bz**2)
            kappa_spitzer = 1e-6 * tg**2.5 # 1e-6 comes from TRAC method (dimensions?)

            L_R = (
                self.emiss_trac_coeff
                * (np.abs(bx) * dx + np.abs(by) * dy + np.abs(bz) * dz)
                / modb
            )

            fmin = np.sqrt(kappa_spitzer * tg / Q_Chianti) / (2.0 * L_R)
            fmin = fmin.where(fmin <= 1, 1.0)

            emis *= fmin
            return emis


    # # # VDEM # # #
    # differential emission measure as a function of temperature and velocity.
    # might also be a function of electron number density (see rcoords_logD_cgs).

    # temperature and velocity bins determined by rcoords_logT and rcoords_vdop_kms:
    cls_behavior_attrs.register('rcoords_vdop_kms')
    rcoords_vdop_kms = simple_property('_rcoords_vdop_kms', default=np.arange(-100,100,10), 
            doc='''velocity [km/s] (or, doppler shift) values, used by some MhdRadiativeLoader vars.
            E.g., used in VDEMs and in getting spectra.
            Can be a scalar, a 1D list or numpy array, or 1D xr.DataArray with dimension 'vdop'.''')
    @known_var
    def get_rcoords_vdop(self):
        '''velocity (or, doppler shift) values used by some vars (e.g. vdems and spectra), in [self.units] system,
        along dimension 'vdop', with coords always in km/s regardless of self.units.
        Determined by self.rcoords_vdop_kms (which can be set directly), converted to [self.units].
        '''
        result = xr1d(self.rcoords_vdop_kms, name='vdop', lenient=True)  # [km/s]
        return result * (self.u("u", convert_from='si') * 1e3)  # [self.units]

    cls_behavior_attrs.register('rcoords_logT')
    rcoords_logT = simple_property('_rcoords_logT', default=np.arange(4.6,6.5,0.1), 
            doc='''log10(Temperature [K]) values used by some MhdRadiativeLoader vars.
            E.g., used in VDEMs and in G(T) lookup tables.
            Can be a scalar, a 1D list or numpy array, or 1D xr.DataArray with dimension 'logT'.
            Note: rcoords_logT minimum value can never be less than self.vdem_logT_min.
                (attempting to set bad value will cause crash & helpful error message.)''')
    @rcoords_logT.setter
    def rcoords_logT(self, value):
        value = xr1d(value, name='logT', lenient=True)
        minval = np.min(value).item()
        if minval < self.vdem_logT_min:
            errmsg = (f'cannot set self.rcoords_logT to something with minimum ({minval:.3g}) less than '
                      f'self.vdem_logT_min={self.vdem_logT_min:.3g}.\n'
                      f'To proceed, first set vdem_logT_min to a smaller value, then try again.\n'
                      f'(To set both simultaneously, use self.vdem_logT = (new_min, new_coords).)')
            raise InputConflictError(errmsg)
        self._rcoords_logT = value

    @known_var
    def get_rcoords_logT(self):
        '''log(Temperature) values used by some vars (e.g. vdems and G(T)), in [self.units] system,
        along dimension 'logT', with coords always telling log(T [K]) regardless of self.units.
        Determined by self.rcoords_logT (which can be set directly), converted to [self.units].
        '''
        result = xr1d(self.rcoords_logT, name='logT', lenient=True)  # log10([K])
        u = self.u("temperature", convert_from='si')
        if u != 1.0:  # [EFF] small efficiency improvement, added because u will very commonly be 1.
            result = result + np.log10(u)   # (this formula also works if u==1, but would just add 0.)
        return result  # log10([self.units])

    cls_behavior_attrs.register('vdem_logT_min', default=4)
    vdem_logT_min = simple_property('_vdem_logT_min', default=4,
            doc='''Minimum log10(temperature [K]) to include in VDEM calculations.
            some VDEMs ignore emiss contributions from very cold regions (T < self.vdem_logT_min).
            Note: vdem_logT_min can never be more than minimum value of self.rcoords_logT.
                (attempting to set bad value will cause crash & helpful error message.)''')
    @vdem_logT_min.setter
    def vdem_logT_min(self, value):
        minval = np.min(self.rcoords_logT).item()
        if value > minval:
            errmsg = (f'cannot set self.vdem_logT_min to something ({value:.3g}) more than '
                      f'the minimum of self.rcoords_logT ({minval:.3g}).\n'
                      f'To proceed, first set rcoords_logT to a larger value, then try again.\n'
                      f'(To set both simultaneously, use self.vdem_logT = (new_min, new_coords).)')
            raise InputConflictError(errmsg)
        self._vdem_logT_min = value

    @property
    def vdem_logT(self):
        '''(self.vdem_logT_min, self.rcoords_logT).
        Can set both at once via self.vdem_logT = (desired_min, desired_coords)
        '''
        return (self.vdem_logT_min, self.rcoords_logT)
    @vdem_logT.setter
    def vdem_logT(self, value):
        newmin, coords = value
        minval = np.min(coords).item()
        if newmin > minval:
            errmsg = (f'cannot set self.vdem_logT = (newmin, newcoords) when newmin > min(newcoords)!\n'
                      f'got newmin={newmin:.3g}, min(newcoords)={minval:.3g}.')
            raise InputConflictError(errmsg)
        # avoid ever having logT_min > min(rcoords_logT):
        oldmin = self.vdem_logT_min
        self.vdem_logT_min = min(oldmin, newmin)
        # actually set the values:
        self.rcoords_logT = coords
        self.vdem_logT_min = newmin  # <-- in case newmin > oldmin

    cls_behavior_attrs.register('rcoords_logD_cgs', default=None)
    rcoords_logD_cgs = simple_property('_rcoords_logD_cgs', default=None, #np.arange(8,11,0.5), 
            doc='''log10(electron number density [cgs units]) values, used by some MhdRadiativeLoader vars.
            E.g., used in vdems and used by G(T,ne) when self.gofnt_mode=='density'.
            Can be a scalar, a 1D list or numpy array, or 1D xr.DataArray with dimension 'logD'.''')
    @known_var
    def get_rcoords_logD(self):
        '''log(electron number density) values for VDEMs and G(T,ne), with ne in [self.units] system,
        along dimension 'logD', with coords always telling log10(ne [cgs]) regardless of self.units.
        Determined by self.rcoords_logD_cgs (which can be set directly), converted to [self.units].

        Getting self('rcoords_logD') will crash if self.rcoords_logD_cgs is None, with a helpful error message.
        '''
        if self.rcoords_logD_cgs is None:
            errmsg = ('Cannot get self("rcoords_logD") because self.rcoords_logD_cgs is None.\n'
                      'To fix, set self.rcoords_logD_cgs to a scalar, a 1D xr.DataArray with dimension "logD"\n'
                      'or a 1D list or numpy array, e.g., self.rcoords_logD_cgs=np.arange(8,11,0.5)')
            raise InputMissingError(errmsg)
        result = xr1d(self.rcoords_logD_cgs, name='logD', lenient=True)  # log10([cgs])
        return result + np.log10(self.u("number_density", convert_from='cgs'))   # log10([self.units])

    @property
    def vdem_varies_logD(self):
        '''tells whether vdems should also vary across logD (electron number density) dimension,
        in addition to temperature (logT) and velocity (vdop) dimensions.
        Currently, equivalent to: (self.rcoords_logD_cgs is not None).
            (Eventually, might allow to edit this property directly, instead.)
        '''
        return self.rcoords_logD_cgs is not None

    # other vdem settings:
    cls_behavior_attrs.register('vdem_ignores_photosphere', default='z=0')
    vdem_ignores_photosphere = simple_property('_vdem_ignores_photosphere', default='z=0',
            doc='''Specifies how the photosphere is treated in VDEM calculations.
            Options include:
                'z=0' --> account for photosphere by ignoring all values at z < 0.
                False --> do not worry about photosphere.''')

    @known_pattern(r'_vdem_ignores_photosphere_(.+)', deps=[0])
    def get__vdem_ignores_photosphere_var(self, var, *, _match=None):
        '''_vdem_ignore_photosphere_var --> self(var) but possibly setting values at z<0 to 0,
        depending on self.vdem_ignores_photosphere.
        '''
        here, = _match.groups()
        arr = self(here)
        if self.vdem_ignores_photosphere == 'z=0':
            arr = arr.where(arr['z'] >= 0, 0)
        return arr

    cls_behavior_attrs.register('vdem_mode', default='interp')
    VDEM_MODE_OPTIONS = {
        'interp': (
            "Interpolate the velocity-differential emission measure (VDEM) in temperature only."
            "This mode uses 'Matthias trick' for temperature binning, providing a balance between speed and accuracy."
        ),
        'nointerp': (
            "No interpolation: "
            "Computes the VDEM by direct binning in both temperature and velocity without any interpolation. "
        ),
        'allinterp': "Interpolate in both temperature and velocity.",
    }
    vdem_mode = simple_property('_vdem_mode', default='interp', validate_from='VDEM_MODE_OPTIONS',
            doc='''Selects the velocity-differential emission measure (VDEM) calculation mode.
            See self.VDEM_MODE_OPTIONS for options.''')

    _DEFAULT_VDEM_LOS = '-component'

    cls_behavior_attrs.register('vdem_los', default=None)
    vdem_los = simple_property('_vdem_los', default=None,
        doc='''tells which line of sight to integrate along for vdem calculations.
        if None, vdems will use self.los if provided, else self._DEFAULT_VDEM_LOS (default: '-component').
            (see self.LOS_OPTIONS for more details about the available shorthand options.
                Another option is to use an xr.DataArray with 'component' dimension.)
        The default ('-component') means the opposite direction implied by self.component,
            e.g. if self.component=='z', then vdem_los is '-z' (corresponding to paths
                which start at top of box and proceed downwards along z).
        Use self.vdem_los_explicit to get the actual value that will be used
            (filling in default or value from self.los as appropriate).''')

    vdem_los_explicit = alias('vdem_los',
        doc='''self.vdem_los, but if None convert to the explicit vdem_los to use.''')
    @vdem_los_explicit.getter
    def vdem_los_explicit(self):
        result = self.vdem_los
        if result is None:
            result = self.los
            if result is None:
                result = self._DEFAULT_VDEM_LOS
        return result

    # vdem vars:
    @known_var(deps=['emiss_for_vdem_loi', 'u_loi_los', 'T_loi', 'rcoords_vdop', 'rcoords_logT'])
    def get_vdem(self):
        '''VDEM (intergrated along an axis), as a function of temperature and velocity.
        If 

        Result depends on self.vdem_mode; see self.VDEM_MODE_OPTIONS for details.
        See also: self.interp_masked

        Regardless of self.vdem_mode, the following always apply:
            temperature and velocity are determined by self.rcoords_logT and self.rcoords_vdop_kms,
                which can be set directly.

            Result is always in [cgs] units, regardless of self.units.
            (Result's coords' units will still be self.units, or self.coords_units if not None.)

            Might ignore photosphere, depending on self.vdem_ignores_photosphere;
                see help(type(self).vdem_ignores_photosphere) for details.

            Result is along line of sight corresponding to self.component (must be a single value).
            Can set via, e.g., self.component='x', or self('vdem_y'), or self(var, component='z').

            Ignores emiss contributions from very cold regions (log10(T) < self.vdem_logT_min).
        '''
        vdem_mode = self.vdem_mode

        with self.using(coords_units=self.coords_units_explicit, units='cgs', los=self.vdem_los_explicit):
            x = self._loi_integration_dim()  # e.g. 'x' if 'component' mode and self.component=='x'.
            # (also, e.g., 'ray_s_dim' or 'ray_s' if in 'ray_draping' mode.)

            ### get relevant vals ###
            dloi = self('dloi')  # get dloi first <--> crashes sooner if crash is needed.
            T = self("T_loi")
            ulos = self('u_loi_los')
            emiss = self('emiss_for_vdem_loi', _T_loi=T)

            if self.vdem_varies_logD:
                ne = self('ne_loi')
                logD = self('rcoords_logD')

            ### compute masks ###
            logT = self('rcoords_logT')
            vdop = self('rcoords_vdop')
            if vdem_mode == 'nointerp':
                maskT = self.interp_masked(np.log10(T), bins=logT, dim=x, mode='simple')
                maskvel = self.interp_masked(ulos, bins=vdop, dim=x, mode='simple')
                arrays = [maskT, maskvel, dloi, emiss]
                if self.vdem_varies_logD:  # put maskD too!
                    maskD = self.interp_masked(np.log10(ne), bins=logD, dim=x, mode='simple')
                    arrays = [maskT, maskvel, maskD, dloi, emiss]
            else:  # nonzero interpolation:
                vel_interp_mode = 'correct' if vdem_mode == 'allinterp' else 'simple'
                maskT, xn = self.interp_masked(np.log10(T), bins=logT, dim=x, return_weights='exp')
                maskvel = self.interp_masked(ulos, bins=vdop, dim=x, mode=vel_interp_mode)

                # apply interpolation correction xn factors to emiss:
                emiss1 = emiss.roll({x:1})
                emiss_interped = emiss * (1.0 - xn) + emiss1 * xn
                arrays = [maskT, maskvel, dloi, emiss_interped]

                if self.vdem_varies_logD:  # put maskD too!
                    maskD = self.interp_masked(np.log10(ne), bins=logD, dim=x, mode=vel_interp_mode)
                    arrays = [maskT, maskvel, maskD, dloi, emiss_interped]

                # drop first point (roll() makes it untrustworthy):
                arrays = [a.isel({x: slice(1, None, None)}) if x in a.dims else a for a in arrays]

            ### compute vdem = (maskT * maskvel * dloi * emiss).sum(x) ###
            vdem = self._vdem_multiply_then_sum_along(*arrays, dim=x)
            return vdem

    @known_var(deps=['(_vdem_ignores_photosphere_emiss)_loi', 'T_loi'])
    def get_emiss_for_vdem_loi(self, *, _T_loi=None):
        '''emiss_for_vdem_loi = emiss_loi, but with some adjustments for vdems:
            - ignores emiss contributions from very cold regions (log10(T) < self.vdem_logT_min).
            - depending on self.vdem_ignores_photosphere, might set to 0 below z=0

        [EFF] for efficiency, can provide _T_loi if known.
        '''
        T = self('T_loi') if _T_loi is None else _T_loi
        emiss = self('(_vdem_ignores_photosphere_emiss)_loi', emiss_opa=False)
        # ^ need to do that before applying tau, because tau depends on loi
        if self.emiss_opa:
            tau = self('tau')  # implied: tau_loi
            emiss = emiss * np.exp(-tau)   # apply tau correction
        emiss = xr.where(T > 10**(self.vdem_logT_min), emiss, 0)
        return emiss

    @known_var(deps=['vdem'], aliases=['vdem_nointerp'])
    def get_vdem_no_interp(self):
        '''VDEM (intergrated along an axis), without interpolation.
        See self.help('vdem') or help(self.get_vdem) for more details about vdems.
        Equivalent: self('vdem', vdem_mode='nointerp').
        '''
        return self('vdem', vdem_mode='nointerp')

    @known_var(deps=['vdem'])
    def get_vdem_interp(self):
        '''VDEM (intergrated along an axis), with interpolation (Matthias trick along T)
        See self.help('vdem') or help(self.get_vdem) for more details about vdems.
        Equivalent: self('vdem', vdem_mode='interp').
        '''
        return self('vdem', vdem_mode='interp')

    @known_var(deps=['vdem'])
    def get_vdem_allinterp(self):
        '''VDEM (intergrated along an axis), with interpolation (Matthias trick along T and u)
        See self.help('vdem') or help(self.get_vdem) for more details about vdems.
        Equivalent: self('vdem', vdem_mode='allinterp').
        '''
        return self('vdem', vdem_mode='allinterp')
    
    # vdem looping details:
    # cls_behavior_attrs <-- intentionally not registered, because shouldn't affect results!
    #  it just affects internal behaviors and efficiency.
    #  instead, added to self._extra_kw_for_quantity_loader_call,
    #    so it can still be adjusted in the usual way, like self(var, vdem_loopdim=...)
    vdem_loopdim = simple_property('_vdem_loopdim', default=('int', 0.45, 'GB'),
        doc=f'''Tells which dimension to loop over during vdems.
        Passed as "loop" kwarg to xarray_multiply_then_sum_along. Docs for that parameter copied below:
        {xarray_multiply_then_sum_along._paramdocs_dict['loop']}''')

    _vdem_errmsg_if_too_big = \
        f'''{xarray_multiply_then_sum_along._errmsg_default}
        
        - for chunking, can adjust self.chunks to break up the problem along self.maindims.
            E.g., self(vdem_var, chunks=dict(x_size=5)) does 5 points at a time along x,
                and joins at the end. It also enables multiprocessing, e.g. self(..., ncpu=10).

        - can also adjust self.vdem_loopdim. E.g. self('vdem', vdem_loopdim='logT')
            Or, vdem_loopdim='int', to loop across the dimension being integrated.
            The default is pretty good too, though: vdem_loopdim=('int', 0.45, 'GB').'''

    def _vdem_multiply_then_sum_along(self, *arrays, dim):
        '''multiply arrays then sum along dim, without making big intermediate array.
        Like xarray_multiply_then_sum_along, but uses vdem-relevant defaults from self.
        '''
        return xarray_multiply_then_sum_along(*arrays, dim=dim, loop=self.vdem_loopdim,
                                              errmsg=self._vdem_errmsg_if_too_big)


    # # # MOMENTS # # #

    @known_pattern(r'(.+)_moments_(.+)', deps=[1])
    def get_moments(self, var, *, _match=None):
        """First three moments of a var, e.g., vdem or spectra. 
        I0 = (int(I du)) [erg/cm^2/s/sr]
        I1 = (int(I u du))/I0 [km/s]
        I2 = (int(I u^2 du))/I0-I1^2 [km/s]

        dim_moments_var --> self(var).pc.moments(dim=dim).
        e.g., vdop_moments_vdem --> self('vdem').pc.moments(dim='vdop')

        result is a Dataset with data_vars 'moment_0', 'moment_1', 'moment_2'.
        """
        dim, var, = _match.groups()
        arr = self(var)
        return xarray_moments(arr, dim=dim)


    # # # Thermal + Instrumental Velocity # # #

    cls_behavior_attrs.register('w_instrumental_si', default=0.0)
    w_instrumental_si = simple_property('_w_instrumental_si', default=0.0,
        doc='''instrumental broadening for thermal velocity, in SI units.
        vtherm_instr = sqrt((kB T / m) + self.w_instrumental_si^2)''')

    @known_var
    def get_w_instrumental(self):
        '''instrumental broadening for thermal velocity, in self.units system.
        w_instrumental = self.w_instrumental_si, converted to self.units system.

        See also: vtherm_instr
        '''
        return xr.DataArray(self.w_instrumental_si * self.u('u', convert_from='si'))

    @known_var(deps=['Tjoule', 'm', 'w_instrumental'], aliases=['vth_instr', 'vthermal_instr'])
    def get_vtherm_instr(self):
        '''thermal velocity including instrumental broadening.
        vtherm_instr = sqrt((kB T / m) + w_instrumental^2)

        w_instrumental can be set via self.w_instrumental_si = value.
        '''
        return (self('Tjoule')  / self('m') + self('w_instrumental')**2)**0.5


    # # # G(T), G(T,n) # # #
    # contribution function(?)
    # depends on electron pressure or number density.

    # electron pressure and number density values determined by rcoords_Pe_cgs and rcoords_logD_cgs.
    # (rcoords_logD_cgs defined above, since it can also be used in vdem.)
    cls_behavior_attrs.register('rcoords_Pe_cgs', default=1e15)
    rcoords_Pe_cgs = simple_property('_rcoords_Pe_cgs', default=1e15, 
            doc='''electron Pressure [cgs units] values, used by some MhdRadiativeLoader vars.
            E.g., used by G(T) when self.gofnt_mode=='pressure'.
            Can be a scalar, a 1D list or numpy array, or 1D xr.DataArray with dimension 'Pe'.''')
    @known_var
    def get_rcoords_Pe(self):
        '''electron Pressure values for G(T) for fixed pressure, in [self.units] system,
        along dimension 'P', with coords always in [cgs] units regardless of self.units.
        Determined by self.rcoords_Pe_cgs (which can be set directly), converted to [self.units].
        '''
        result = xr1d(self.rcoords_Pe_cgs, name='Pe', lenient=True)  # [cgs]
        return result * self.u("pressure", convert_from='cgs')  # [self.units]

    # other G(T) settings:
    cls_behavior_attrs.register('gofnt_mode', default='pressure')
    GOFNT_MODE_OPTIONS = {'pressure': 'Compute G(T) for a fixed electron pressure',
                          'density': 'Compute G(T,ne)'
                          }
    gofnt_mode = simple_property('_gofnt_mode', default='pressure', validate_from='GOFNT_MODE_OPTIONS',
            doc='''Selects the G(T) calculation mode. See self.GOFNT_MODE_OPTIONS for options.''')
    
    cls_behavior_attrs.register('gofnt_interp_mode', default='interp')
    GOFNT_INTERP_MODE_OPTIONS = {
        'interp': (
            "Interpolate the emission measure x G(T) in temperature."
            "This mode uses 'Matthias trick' for temperature binning, providing a balance between speed and accuracy."
        ),
        'nointerp': (
            "No interpolation: "
            "Computes the emission measure x G(T) direct binning in both temperature and velocity without any interpolation. "
        ),
    }
    gofnt_interp_mode = simple_property('_gofnt_interp_mode', default='interp',
            validate_from='GOFNT_INTERP_MODE_OPTIONS',
            doc='''Selects the emission measure x G(T) calculation mode.
            See self.GOFNT_INTERP_MODE_OPTIONS for options.''')

    cls_behavior_attrs.register('gofnt_abundance', default='sun_coronal_2021_chianti')
    gofnt_abundance = simple_property('_gofnt_abundance', default='sun_coronal_2021_chianti',
            doc='''Abundance(?) to use for G(T) calculations.
            Passed directly to muse.instr.utils.chianti_gofnt_linelist via "abundance" kwarg.''')

    cls_behavior_attrs.register('gofnt_wavelength_range', default=[171.07, 171.08])
    gofnt_wavelength_range = simple_property('_gofnt_wavelength_range', default=[171.07, 171.08],
        doc='''Wavelength range (as [min, max]) (in [Angstrom] units) to use for G(T) calculations.
        Passed directly to muse.instr.utils.chianti_gofnt_linelist via "wavelength_range" kwarg.''')

    cls_behavior_attrs.register('rcoords_wavelength_A', default=171.07)
    rcoords_wavelength_A = simple_property('_rcoords_wavelength_A', default=171.07,
        doc='''Wavelength (in [Angstrom] units) used by some MhdRadiativeLoader vars.
        E.g., used in get_spectra().
        [TODO] can it be a 1D list/array, or is it required to be a scalar?''')
    
    
    # G(T) vars
    @known_pattern(r'GofT_(.+)_ion', deps=['rcoords_logT'],
                   attr_deps=[('gofnt_mode', {'pressure': 'rcoords_Pe', 'density': 'rcoords_logD'})])
    def get_Gofnt(self, var, *, _match=None):
        '''G(T,ne) from ChiantiPy and muse library for a given ion. E.g., GofT_fe_9_ion
        Always in [cgs] (erg*cm^3/s/sr) regardless of self.units.

        ion string (like 'fe_9') is passed into muse.instr.utils.chianti_gofnt_linelist,
            via ionList = [ionstr]. See that function for more details about available ions.

        "returns G(T,ne) and spectral line properties, including wavelength and transition probabilities." 
        [TODO] what does that mean?

        CAUTION: directly adjusts the value of self.rcoords_wavelength_A, based on result.
        '''
        ionstr, = _match.groups()

        # finds in Chiantypy the line list with the properties listed above
        if self.gofnt_mode == "pressure":
            line_list = muse_instr_utils.chianti_gofnt_linelist(
                temperature = 10**self('rcoords_logT'),
                pressure = self('rcoords_Pe', units='cgs'),  # ChiantiPy wants cgs
                abundance = self.gofnt_abundance,
                wavelength_range = self.gofnt_wavelength_range,
                ionList = [ionstr],
                minimum_abundance = None,
            )

            # Selecting the strongest line
            line_list["gofnt_max"] = line_list.gofnt.sum(['logT']) 
        elif self.gofnt_mode == 'density':
            line_list = muse_instr_utils.chianti_gofnt_linelist(
                temperature = 10**self('rcoords_logT'),
                density = 10**self('rcoords_logD', units='cgs'),   # ChiantiPy wants cgs
                abundance = self.gofnt_abundance,
                wavelength_range = self.gofnt_wavelength_range,
                ionList = [ionstr],
                minimum_abundance = None,
            )

            # Selecting the strongest line
            line_list["gofnt_max"] = line_list.gofnt.sum(['logT']) 
        else:
            assert False, f"coding error if reached this line; invalid gofnt_mode: {self.gofnt_mode!r}"
        sort_index = np.argsort(-line_list.gofnt_max, 
                                axis=line_list.gofnt_max.get_axis_num('trans_index'))
        line_list_sort = line_list[dict(trans_index=sort_index)]
        line_list_sort_c = line_list_sort.isel(trans_index=0)
        # SE - it's weird for a known_var to alter a behavior_attr
        # JMS yes ... but it's convenient here.
        # [TODO] long-term, better to include directly in output somehow (e.g., array.attrs?)
        #    (SE - for now it is okay. left a note in docstring to let user know the attr will be adjusted.)
        self.rcoords_wavelength_A = line_list_sort_c.squeeze().wvl.values
        return line_list_sort_c

    @known_pattern(r'GofT_(.+)_ion_atmos', deps=['T', {0: 'GofT_{group0}_ion'}],
                   attr_deps=[('gofnt_mode', {'density': 'ne'})])
    def get_Gofnt_atmos(self, var, *, _match=None):
        '''Interpolated G(T) with temperature, for a given ion.
        Always in [cgs] (erg*cm^3/s/sr) regardless of self.units.
        For details about available ions, see muse.instr.utils.chianti_gofnt_linelist
        '''
        ionstr, = _match.groups()
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):
            goftn = self(f'GofT_{ionstr}_ion').gofnt * self.emiss_norm
            tg = self('T')
            if self.gofnt_mode == "pressure":
                return goftn.interp(logT=np.log10(tg))
            elif self.gofnt_mode == "density": 
                nel = self('ne')
                return goftn.interp(logT=np.log10(tg),logD=np.log10(nel))
            else:
                assert False, f"coding error if reached this line; invalid gofnt_mode: {self.gofnt_mode!r}"

    @known_pattern(r'GofT_(.+)_ion_em',
                   attr_deps=[('gofnt_interp_mode', {'interp': [{0: 'GofT_{group0}_ion_em_interp'}],
                                                   'nointerp': [{0: 'GofT_{group0}_ion_em_no_interp'}]})])
    def get_Gofnt_em(self, var, *, _match=None):
        '''Emission * G(T) for a given ion. Results always in [cgs] (erg/cm^2/s/sr).
        For details about available ions, see muse.instr.utils.chianti_gofnt_linelist
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        ionstr, = _match.groups()

        if self.gofnt_interp_mode == "interp": 
            return self(f"GofT_{ionstr}_ion_em_interp")
        elif self.gofnt_interp_mode == "nointerp": 
            return self(f"GofT_{ionstr}_ion_em_no_interp")
        else:
            assert False, f"coding error if reached this line; invalid gofnt_interp_mode: {self.gofnt_interp_mode!r}"

    @known_pattern(r'GofT_(.+)_ion_em_no_interp', deps=['emiss', {0: 'GofT_{group0}_ion_atmos'}])
    def get_Gofnt_em_nointerp(self, var, *, _match=None):
        '''Emission * G(T) for a given ion. Results always in [cgs] (erg/cm^2/s/sr).
        For details about available ions, see muse.instr.utils.chianti_gofnt_linelist
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        ionstr, = _match.groups()
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):      
            goftn = self(f'GofT_{ionstr}_ion_atmos')
            em = self('emiss')
            return goftn * em

    @known_pattern(r'GofT_(.+)_ion_em_interp', deps=['emiss', 'T', {0: 'GofT_{group0}_ion'}])
    def get_Gofnt_em_interp(self, var, *, _match=None): # uhm .... not sure why, but for not sharp temperature gradient has an artifact ... 
        '''Emission * G(T) for a given ion. Results always in [cgs] (erg/cm^2/s/sr).
        For details about available ions, see muse.instr.utils.chianti_gofnt_linelist
        '''
        self._radiative_assert_single_component()
        ionstr, = _match.groups()
        x = str(self.component)
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):      
            goftn_atmos = self(f'GofT_{ionstr}_ion_atmos')
            emiss = self('emiss')
            emiss = emiss * (emiss["z"] > 0)  # [TODO] connect to "vdem_ignores_photosphere" option or something like it

            emiss1 = emiss.roll({x:1})
            # removing emission from first cell due to roll
            emiss1[{x:0}] = 0.0

            masktg, xn = self.interp_masked(np.log10(self('T')), bins=self('rcoords_logT'), dim=x, return_weights='exp')
            emiss_interped = emiss * (1.0 - xn) + emiss1 * xn  # emiss with interp correction weighting adjacent cells

            result = masktg * goftn_atmos * emiss_interped
            return result.sum(dim='logT')


    @known_var(deps=['ne', 'T', 'r', 'dloi'])
    def get_tau_500(self, wavelength=None):
        '''Computes the optical depth at 500 nm.
        '''

        with self.using(coords_units=self.coords_units_explicit, units='cgs', los=self.vdem_los_explicit):   

            with self.using(coords_units=self.coords_units_explicit, units='cgs'):      
                x = self._loi_integration_dim()  # e.g. 'x' if 'component' mode and self.component=='x'.
                # (also, e.g., 'ray_s_dim' or 'ray_s' if in 'ray_draping' mode.)

                ### get relevant vals ###
                dloi = self('dloi') 

                kb = DEFAULTS.PHYSICAL.CONSTANTS_SI['kB'] * 1e7  # cgs units
                grph = (self.elements.n_per_nH() * self.elements.m * self.u('amu')).sum() 
                const = (1.03526e-16 / grph) * 2.9256e-17
                ev_to_erg = 1.60218e-12
                tg = self('T_loi')
                xhmbf = const * self('ne_loi') / tg**1.5 * np.exp(0.754e0 * ev_to_erg / kb / tg) * self('r_loi')
                tau = self.loi_cumsum(xhmbf * dloi, do_interp=False)
                return tau
        
    @known_var
    def get_opac_h_he(self, wavelength=None):
        '''Opacities at wavelength=self.rcoords_wavelength_A.
        from hydrogen and helium bound free absorption (Anzer & Heinzel, 2005)
        '''
        rhe = 0.1
        opa1d = self('chianti_opa1d_table')
        arr = (opa1d.h) * self('opac_h') + rhe * ((1 - opa1d.hei - (1-opa1d.hei-opa1d.he)) *
                                                  self('opac_hei') + (opa1d.hei) * self('opac_heii'))
        arr[arr < 0] = 0
        return arr

    @known_var
    def get_chianti_opa1d_table(self, tabname='chianti'):
        '''Ionization state table from ChiantiPy.
        result is an xr.Dataset with data_vars 'h', 'he', 'hei',
            telling the H, HeI, and HeII ionization fractions, respectively.
        '''
        h = ch.Ioneq.ioneq(1)
        h.load(tabname)
        he = ch.Ioneq.ioneq(2)
        he.load(tabname)
        opa1d = xr.Dataset()
        opa1d['he'] = xr.DataArray(he.Ioneq[0, :], coords={"logT": np.log10(he.Temperature)}, dims="logT")
        opa1d['hei'] = xr.DataArray(he.Ioneq[1, :], coords={"logT": np.log10(he.Temperature)}, dims="logT")
        opa1d['h'] = xr.DataArray(h.Ioneq[0, :], coords={"logT": np.log10(h.Temperature)}, dims="logT")
        return opa1d

    @known_var
    def get_opac_h(self):
        '''H opacity at self.rcoords_wavelength_A.
        See Anzer & Heinzel, 2005 for details. 
        '''
        ghi = 0.99
        o0 = 7.91e-18  # cm^2
        ohi = 0
        if self.rcoords_wavelength_A <= 912: # Amstrong
            ohi = o0 * ghi * (self.rcoords_wavelength_A / 912.0)**3
        return ohi

    @known_var
    def get_opac_hei(self):
        '''HeI opacity at self.rcoords_wavelength_A.
        See Anzer & Heinzel, 2005 for details. 
        '''
        c = [-2.953607e1, 7.083061e0, 8.678646e-1,
             -1.221932e0, 4.052997e-2, 1.317109e-1,
             -3.265795e-2, 2.500933e-3]
        ohei = 0
        if self.rcoords_wavelength_A <= 504: # Amstrong
            for i, cf in enumerate(c):
                ohei += cf * (np.log10(self.rcoords_wavelength_A))**i
            ohei = 10.0**ohei
        return ohei

    @known_var
    def get_opac_heii(self):
        '''HeII opacity at self.rcoords_wavelength_A.
        See Anzer & Heinzel, 2005 for details. 
        '''
        gheii = 0.85
        o0 = 7.91e-18  # cm^2
        oheii = 0
        if self.rcoords_wavelength_A <= 228: # Amstrong
            oheii = 16 * o0 * gheii * (self.rcoords_wavelength_A / 912.0)**3
        return oheii

    @known_var(deps=['r', 'opac_h_he', 'T'])
    def get_absorption_h_he(self):
        '''Absorption coefficient at wavelength=self.rcoords_wavelength_A,
            considering only H, HeI, and HeII contributions.
        Result is always in [cgs] units, regardless of self.units.
        See Anzer & Heinzel, 2005 for details.

        absorption_h_he = (opacity from h and he) * (single-fluid mass density) / grph,
            where grph = mean molecular mass (same unit system as other terms in expression).
            Assuming photopsheric abundances, grph is roughly 1.3 to 1.5 amu, depending on
            exact abundances model (but constant in space & time; it just depends on self.elements).
        '''
        with self.using(coords_units=self.coords_units_explicit, units='cgs'):   
            rho = self("r")
            grph = (self.elements.n_per_nH() * self.elements.m * self.u('amu')).sum() 
            opac_h_he = self('opac_h_he')
            opac = opac_h_he.interp(logT=np.log10(self('T')))
            opac = opac.fillna(opac.max())
            return (opac * rho) / grph

    @known_var(deps=['absorption_h_he_loi', 'dloi'])
    def get_tau(self):
        '''Optical depth along line of integration (default: loi = line of sight),
        at wavelength=self.rcoords_wavelength_A, and considering only H, HeI, and HeII contributions.
        Result is always in [cgs] units, regardless of self.units.

        loi determined by self.loi (see help(type(self).loi) for details),
            when using self.los = self.vdem_los_explicit (see help(type(self).vdem_los) for details).

        tau(l) = integral(absorption * dl, from 0th l value to l); use self.loi instead of l.
            (in python terms, this is represented by a cumsum.)

        To reverse direction (i.e. "start from the other end"), edit self.loi or self.vdem_los.
            E.g. when self.loi='los', swapping from self.vdem_los='-x' to self.vdem_los='x' swaps direction.
        '''
        with self.using(coords_units=self.coords_units_explicit, units='cgs', los=self.vdem_los_explicit):   
            dloi = self('dloi')   # get dloi first <--> crash sooner if crash is needed.
            absorption = self('absorption_h_he_loi')
            result = self.loi_cumsum(absorption * dloi, do_interp=False)  # interp already handled above
            return result

    @known_pattern(r'spectra_(.+)_ion_prof',
                   deps=['vtherm_instr', 'u', 'rcoords_vdop', {0: 'GofT_{group0}_ion_em'}],
                   attr_deps=[('emiss_mode', {'notrac_opa': 'tau', 'trac_opa': 'tau'})])
    def get_spectra(self, var, *, _match=None): # this one could be break in two functions, one with and the other without abs. 
        '''Synthetic spectra for a given ion. Always in [cgs] units, regardless of self.units.
        Units erg/cm^2/s/sr/Angs

        Result depends on self.rcoords_vdop_kms and self.rcoords_wavelength_A, which can be set directly.
        Result varies along the 'vdop' dimension which indicates doppler velocity (in km/s).

        [TODO] formula/details
        '''
        # [TODO] handle loi properly. For now, might be okay with simple loi, but will crash if ray_draping.
        ionstr, = _match.groups()
        self._radiative_assert_single_component()
        with self.using(coords_units=self.coords_units_explicit, units='cgs', los=self.vdem_los_explicit):
            # force noopa emiss mode, when getting _ion_em
            emiss_mode = self.emiss_mode
            if "_opa" in emiss_mode:
                emiss_mode = emiss_mode.replace("_opa", "_noopa")
            em = self(f'GofT_{ionstr}_ion_em', emiss_mode=emiss_mode)
            sig = self('vtherm_instr')
            vlos = self('u_los')  # [TODO] u_loi_los
            vdop = self('rcoords_vdop')
            wvl = doppler_to_lambda(self.rcoords_vdop_kms, self.rcoords_wavelength_A)  # from km/s to Angstrom
            with self.maintaining('wavelength'):
                I_slice = []
                for i_v, v_arr in enumerate(vdop):  # looping because "tau" not is ready to handle 'vdop' dim.
                    if "_opa" in self.emiss_mode:
                        self.rcoords_wavelength_A = wvl[i_v]
                        tau = self('tau')
                        xarray_result_size_check(tau, em)  # prevent wayyy too big result.
                        em_opa = em * np.exp(-self("tau"))
                    else:
                        em_opa = em
                    phi = np.exp(-0.5 * ((v_arr - vlos) / sig)**2) / (np.sqrt(2 * np.pi) * sig)
                    dlos = self('dloi')
                    xarray_result_size_check(em_opa, phi, dlos)  # prevent wayyy too big result.
                    I_slice.append((em_opa * phi * dlos).sum(dim=los))
            int = xr.concat(I_slice,dim="vdop")
            int = int.assign_coords(wavelength=("vdop",wvl))
        return int


    # # # INTERPOLATION CORRECTIONS # # #

    @format_docstring(_static_helper_doc=interp_masked.__doc__, sub_ntab=2)
    def interp_masked(self, array_or_var, bins, dim=None, *, mode='correct', return_weights=False):
        """
        Compute interpolation masks and normalization factors for a variable (e.g., "T")
        for use in DEM/VDEM binning with or without logarithmic scaling.

        Works just like the "static" method (not attached to this class),
        but with a few extra convenient features here:
            (1) first arg can be array (as in static method), OR a var.
                if it is a var (i.e., a str), use array=self(var)
            (2) dim does not need to be provided;
                if not provided, use dim=str(self.component).

        Docstring from "static" method copied below, for convenience.
        -------------------------------------------------------------
            {_static_helper_doc}
        """
        if isinstance(array_or_var, str):
            var = array_or_var
            array = self(var)
        else:
            array = array_or_var
        if dim is None:
            dim = str(self.component)
        return interp_masked(array=array, bins=bins, dim=dim, mode=mode, return_weights=return_weights)


    # # # VDEM PIPELINE # # #

    def vdem_pipeline(self, *, 
                      los_dim='z', 
                      iz0 = 50, 
                      tg_percent = 0.1, 
                      dlogT = 0.1,
                      mintg_cut = 4.6, 
                      maxtg_cut = 7.7, 
                      tg_bins = 60,
                      ulos_bin = 60,
                      vel_percent = 0.1,
                      dvdop = 10,
                      minvel_cut = -3000,
                      maxvel_cut = 3000,
                      emiss_norm=1e27,
                      modelname = '',
                      zarr_version = 2,
                      chunks=256,
                      ncpu=12,
                      dst='pc_vdem_pipeline',
                      author,
                      ):
        """
        Execute a standard VDEM (Velocity-Differential Emission Measure) analysis pipeline.

        This pipeline performs a complete VDEM analysis including:
        1. Automatic determination of temperature and velocity ranges from data histograms
        2. Computation of VDEM with configurable binning
        3. Generation of moment plots (intensity, mean velocity, width)
        4. Photosphere region analysis with separate plots
        5. Saving results to Zarr format with complete metadata

        The pipeline automatically determines optimal temperature and velocity bins based on
        histogram analysis, applies user-specified cutoffs, computes VDEM along the specified
        line-of-sight dimension, and generates diagnostic plots.

        Parameters
        ----------
        los_dim : str, default='z'
            Line-of-sight dimension for VDEM integration ('x', 'y', or 'z').
        iz0 : int, default=50
            Index along z-dimension marking the photosphere boundary. Used to define
            integration limits and photosphere cutoff analysis.
        tg_percent : float, default=0.1
            Percentage threshold for automatic temperature range determination.
            Histogram bins with density > tg_percent% of maximum are included.
        dlogT : float, default=0.1
            Temperature bin width in log10(T [K]) units.
        mintg_cut : float or None, default=4.6
            Minimum log10(temperature [K]) cutoff. Overrides automatic determination
            if specified.
        maxtg_cut : float or None, default=7.7
            Maximum log10(temperature [K]) cutoff. Overrides automatic determination
            if specified.
        tg_bins : int, default=60
            Number of bins for temperature histogram analysis.
        ulos_bin : int, default=60
            Number of bins for velocity histogram analysis.
        vel_percent : float, default=0.1
            Percentage threshold for automatic velocity range determination.
            Histogram bins with density > vel_percent% of maximum are included.
        dvdop : float, default=10
            Velocity bin width in km/s.
        minvel_cut : float or None, default=-3000
            Minimum velocity cutoff in km/s. Overrides automatic determination if specified.
        maxvel_cut : float or None, default=3000
            Maximum velocity cutoff in km/s. Overrides automatic determination if specified.
        emiss_norm : float, default=1e27
            Emission measure normalization factor.
        modelname : str, default=''
            Model identifier string used in output filenames.
        zarr_version : int, default=2
            Zarr format version for output files (2 or 3).
        chunks : int, default=256
            Chunk size for parallel computation along spatial dimensions.
        ncpu : int, default=12
            Number of CPU cores to use for parallel computation.
        dst : str, default='pc_vdem_pipeline'
            Destination directory for output files.
        author : str, required
            Author name/identifier for output filenames and metadata.

        Returns
        -------
        vdem0 : xarray.Dataset
            Processed VDEM dataset with coordinates and attributes set for the full domain.
            Contains 'vdem' data variable with dimensions (logT, vdop, x, y).

        Notes
        -----
        The pipeline generates the following outputs:
        
        - Temperature and velocity histogram plot: 
          'T_ulos_hist_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM moments plot (full domain): 
          'vdem_moment_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM moments plot (photosphere): 
          'vdem_moment_ph_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM Zarr dataset: 
          'vdem_{code}_{modelname}_los{los_dim}_{author}_{date}_{snap}.zarr'

        The pipeline automatically:
        - Sets units to CGS
        - Determines integration slice from iz0 to top of domain
        - Computes histograms to find optimal binning
        - Applies both automatic and manual cutoffs for T and velocity ranges
        - Computes VDEM with multiprocessing support
        - Generates moment plots with logarithmic scaling for intensity
        - Performs separate photosphere analysis (iz0 to iz0+5)
        - Adds complete history and metadata to output

        Examples
        --------
        >>> # Standard VDEM pipeline with default settings
        >>> result = calc.vdem_pipeline(author='scientist_name')
        
        >>> # Custom temperature and velocity ranges
        >>> result = calc.vdem_pipeline(
        ...     mintg_cut=5.0,
        ...     maxtg_cut=7.0,
        ...     minvel_cut=-1000,
        ...     maxvel_cut=1000,
        ...     author='scientist_name'
        ... )
        
        >>> # High-resolution analysis with more CPU cores
        >>> result = calc.vdem_pipeline(
        ...     dlogT=0.05,
        ...     dvdop=5,
        ...     chunks=128,
        ...     ncpu=24,
        ...     author='scientist_name'
        ... )

        See Also
        --------
        limb_vdem_pipeline : Pipeline for limb observations with ray draping
        get_vdem : Core VDEM computation method
        """

        import matplotlib.pyplot as plt
        from matplotlib import colors
        from dateutil.tz import gettz
        try:
            import zarr
        except ImportError as err:
            errmsg = ("vdem_pipeline requires zarr; try again after installing, e.g. 'pip install zarr'.")
            raise ImportError(errmsg) from err

        WEST = gettz("US/Western")

        chosen_code = type(self).__name__.replace("Calculator","")
        
        self.units = 'cgs'
        self.emiss_norm = emiss_norm
        self.component = los_dim
        top = np.argmax(self.get_maindims_coords()['z'])
        if iz0 is None:
            iz0 = np.argmin(np.abs(self.get_maindims_coords()['z']))
        if top < iz0:
            sl = slice(top,iz0)
        else:
            sl = slice(iz0,top)

        self.slices = ({'z':sl})
        tg = self("T")
        ulos = self("u", units = 'si') / 1e3 # to km/s

        # The histogram allows to find the temperature and velocity limits
        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        hist_values = np.log10(tg).plot.hist(density=True, bins=tg_bins, ax=ax[0])

        maxtg = (
                np.ceil(
                    np.max(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * tg_percent / 100.0)])
                    / dlogT
                )
                * dlogT
            )
        mintg = (
                np.floor(
                    np.min(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * tg_percent / 100.0)])
                    / dlogT
                )
                * dlogT
            )
        if mintg_cut is not None:
            mintg = np.max([mintg, mintg_cut])
        if maxtg_cut is not None:
            maxtg = np.min([maxtg, maxtg_cut])

        hist_values = ulos.plot.hist(density=True, bins=ulos_bin, ax=ax[1])
        fig.tight_layout()
        saveto = f"T_ulos_hist_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        os.makedirs(dst, exist_ok=True)
        fig.savefig(os.path.join(dst, saveto))
        
        maxvel = (
                np.ceil(
                    np.max(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * vel_percent / 100.0)])
                    / dvdop
                )
                * dvdop
            )
        minvel = (
                np.floor(
                    np.min(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * vel_percent / 100.0)])
                    / dvdop
                )
                * dvdop
            )
        if minvel_cut is not None:
            minvel = np.max([minvel, minvel_cut])
        if maxvel_cut is not None:
            maxvel = np.min([maxvel, maxvel_cut])
        
        self.rcoords_logT = np.arange(mintg, maxtg, dlogT)
        self.rcoords_vdop_kms = np.arange(minvel, maxvel, dvdop)
        self.tabin.extrapolate_kind = "constant"
        vdem = self('vdem', chunks=dict(x=chunks), ncpu=ncpu)
        if self.vdem_varies_logD:
            vdem_sumd = vdem.sum(dim='logD')
        else: 
            vdem_sumd = vdem
        mom = vdem_sumd.pc.moments(dim='vdop')
        fig, ax = plt.subplots(1, 3, figsize=(12, 6))
        mom.moment_0.sum(dim='logT').plot.imshow(norm=colors.LogNorm(vmin=1e-5,vmax=1e5), ax=ax[0])
        mom.moment_1.sum(dim='logT').plot.imshow(ax=ax[1], cmap="bwr")
        mom.moment_2.sum(dim='logT').plot.imshow(ax=ax[2])
        fig.tight_layout()
        saveto = f"vdem_moment_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        fig.savefig(os.path.join(dst, saveto))

        vdem_cut = self('vdem', 
                        slices=dict(z=slice(iz0, iz0+5)),
                        emiss_mode = 'notrac_noopa',
                        vdem_logT = (0, np.arange(3.0,4.5,0.1)),
                        rcoords_vdop_kms=np.arange(-1.0e1,1.e1,1),
                        rcoords_logD_cgs = None, 
                        vdem_mode='nointerp',
                        vdem_ignores_photosphere=False,
                        vdem_loopdim=None,
                        chunks=dict(x=chunks),
                        ncpu=ncpu,
                        component='z',
                        )

        mom = vdem_cut.pc.moments(dim='vdop')
        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        mom.moment_0.sum(dim='logT').plot(ax=ax[0])
        mom.moment_1.sum(dim='logT').plot(vmin=-5,vmax=5,cmap='bwr', ax=ax[1])
        ax[1].set_title("Doppler velocity (positive must be downward)")
        fig.tight_layout()
        saveto = f"vdem_moment_ph_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        fig.savefig(os.path.join(dst, saveto))

        xarray_update_call_history(vdem, locals(), self.vdem_pipeline)

        vdem0 = vdem.to_dataset(name='vdem')
        # _xarray_save_prep(vdem)[0] trashes PlasmaCalcs rich metadata like Snap and Component objects,
        #  and adds some helpful attrs like pc__version__, datetime, and maybe pc__commit_hash.
        vdem0 = _xarray_save_prep(vdem0, add_code_snapshot_info=dict(default='_undefined_'))[0]
        vdem0.x.attrs["long_name"] = "X"
        vdem0.y.attrs["long_name"] = "Y"
        vdem0.x.attrs["units"] = "cm"
        vdem0.y.attrs["units"] = "cm"
        vdem0.vdem.attrs["units"] = "1e27 / cm5"
        vdem0.vdem.attrs["description"] = "DEM(T,vel,x,y)"
        vdem0.vdop.attrs["long_name"] = r"v$_{Doppler}$"
        vdem0.vdop.attrs["units"] = "km/s"
        vdem0.logT.attrs["long_name"] = r"log$_{10}$(T)"
        vdem0.logT.attrs["units"] = r"log$_{10}$ (K)"
        saveto = f"vdem_{chosen_code}{modelname}_los{los_dim}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.zarr",
        vdem0.to_zarr(os.path.join(dst, saveto), zarr_version=zarr_version)
        return vdem0

    def limb_vdem_pipeline(self, *, 
                      iz0 = 50,
                      snap = None,
                      tile_snaps = None,
                      fov_method = "standard",
                      Abox_x_NXEXT = 0,
                      Nx_times = 1,
                      Ny_times = 1,
                      Nz_times = 1,
                      Asph_theta=np.pi/2,
                      Asph_phi=0,
                      tg_percent = 0.1, 
                      dlogT = 0.1,
                      mintg_cut = 4.6, 
                      maxtg_cut = 7.7, 
                      tg_bins = 60,
                      ulos_bin = 60,
                      vel_percent = 0.1,
                      dvdop = 10,
                      minvel_cut = -3000,
                      maxvel_cut = 3000,
                      emiss_norm = 1e27,
                      modelname = '',
                      nslicer = 46,
                      zarr_version = 2,
                      dst = 'pc_limb_vdem_pipeline',
                      author,
                      ):
        """
        Execute a VDEM pipeline for limb observations using ray draping geometry.

        This specialized pipeline performs VDEM analysis for solar limb observations where
        the line-of-sight follows curved ray paths through the simulation domain. It uses
        ray draping to simulate realistic observing geometries and supports tiling multiple
        snapshots for extended spatial coverage.

        The pipeline handles large datasets by computing VDEM in slices, performs both
        full-disk and photosphere-cutoff analyses, and generates diagnostic plots with
        proper coordinate transformations for limb viewing geometry.

        Parameters
        ----------
        iz0 : int, default=50
            Index along z-dimension marking the photosphere boundary. Used for
            photosphere cutoff analysis.
        snap : int or None, optional
            Snapshot number to analyze. If None, uses current snapshot.
        tile_snaps : array-like or None, optional
            Array of snapshot indices to tile along x-dimension for extended coverage.
            E.g., [0, 1, 2, 'current'] tiles three consecutive snapshots, before the current snapshot.
            ('current' or 'here' must appear in tile_snaps exactly once.)
            Note that the full (tiled) extent will be repeated periodically as needed to cover rays.
        fov_method : str, default='standard'
            Field-of-view generation method. Options depend on ray draping implementation.
        Abox_x_NXEXT : float, default=0
            X-offset for ray draping box in units of simulation XEXT (domain extent).
            Allows shifting the observing geometry along x.
        Nx_times : float, default=1
            Multiplier for number of FOV pixels in x-direction relative to simulation Nx.
        Ny_times : float, default=1
            Multiplier for number of FOV pixels in y-direction relative to simulation Ny.
        Nz_times : float, default=1
            Multiplier for number of ray integration points relative to simulation Nz.
        Asph_theta : float, default=np.pi/2
            Spherical coordinate theta (colatitude) angle in radians for viewing geometry.
            pi/2 corresponds to equatorial view.
        Asph_phi : float, default=0
            Spherical coordinate phi (azimuthal) angle in radians for viewing geometry.
        tg_percent : float, default=0.1
            Percentage threshold for automatic temperature range determination from histogram.
        dlogT : float, default=0.1
            Temperature bin width in log10(T [K]) units.
        mintg_cut : float or None, default=4.6
            Minimum log10(temperature [K]) cutoff. Overrides automatic determination.
        maxtg_cut : float or None, default=7.7
            Maximum log10(temperature [K]) cutoff. Overrides automatic determination.
        tg_bins : int, default=60
            Number of bins for temperature histogram analysis.
        ulos_bin : int, default=60
            Number of bins for velocity histogram analysis.
        vel_percent : float, default=0.1
            Percentage threshold for automatic velocity range determination from histogram.
        dvdop : float, default=10
            Velocity bin width in km/s.
        minvel_cut : float or None, default=-3000
            Minimum velocity cutoff in km/s. Overrides automatic determination.
        maxvel_cut : float or None, default=3000
            Maximum velocity cutoff in km/s. Overrides automatic determination.
        emiss_norm : float, default=1e27
            Emission measure normalization factor.
        modelname : str, default=''
            Model identifier string used in output filenames.
        nslicer : int, default=46
            Number of slices for dividing FOV computation. Higher values reduce memory
            usage but increase I/O overhead.
        zarr_version : int, default=2
            Zarr format version for output files (2 or 3).
        dst : str, default='pc_limb_vdem_pipeline'
            Destination directory for output files.
        author : str, required
            Author name/identifier for output filenames and metadata.

        Returns
        -------
        vdem0 : xarray.Dataset
            Processed VDEM dataset with limb observation coordinates and complete metadata.
            Contains 'vdem' data variable with dimensions (logT, vdop, x, y) where x and y
            represent FOV coordinates.

        Notes
        -----
        **Temporary Directories**: This pipeline creates and DELETES temporary directories:
        - '_pc_temp_fov_lib_vdem/': For storing partial VDEM computations
        - '_pc_temp_fov_lib_vdem_ph/': For photosphere-cutoff VDEM computations
        
        These directories will be destroyed via shutil.rmtree() during pipeline execution.

        **Ray Draping Setup**: The pipeline configures ray draping with:
        - R0 = Solar radius (from DEFAULTS.PHYSICAL.CONSTANTS_SI)
        - T = 1 AU (observer distance)
        - Integration points scaled by Nz_times
        - FOV dimensions scaled by Nx_times
        - Spherical viewing angles (Asph_theta, Asph_phi)

        **Output Files**:
        - Temperature/velocity histograms: 
          'T_ulos_hist_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM moments (full): 
          'vdem_moment_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM moments (photosphere): 
          'vdem_moment_ph_{code}_{modelname}_{author}_{date}_{snap}.png'
        - VDEM Zarr dataset: 
          'vdem_{code}_{modelname}_limb_{author}_{date}_snaps_{tiles}.zarr'

        **Processing Steps**:
        1. Configure ray draping geometry with spherical coordinates
        2. Enable snapshot tiling if tile_snaps provided
        3. Compute histograms to determine optimal T and velocity ranges
        4. Compute VDEM in FOV slices to manage memory
        5. Merge slices and compute moments for full domain
        6. Repeat with photosphere cutoff (z < iz0-15)
        7. Generate diagnostic plots with proper scaling
        8. Save with complete metadata and coordinate attributes

        Examples
        --------
        >>> # Standard limb observation at solar equator
        >>> result = calc.limb_vdem_pipeline(
        ...     snap=100,
        ...     tile_snaps=[99, 'current', 101],
        ...     author='scientist_name'
        ... )
        
        >>> # High-resolution polar limb observation
        >>> result = calc.limb_vdem_pipeline(
        ...     Asph_phi=-np.pi/2,  # North pole view
        ...     Nx_times=2,    # Double resolution
        ...     dlogT=0.05,    # Finer temperature bins
        ...     nslicer=92,    # More slices for memory
        ...     author='scientist_name'
        ... )
        
        >>> # Limb observation with shifted viewpoint
        >>> result = calc.limb_vdem_pipeline(
        ...     Abox_x_NXEXT=0.5,  # Shift by half domain width
        ...     Asph_phi=np.pi/4,  # 45-degree azimuthal angle
        ...     author='scientist_name'
        ... )

        See Also
        --------
        vdem_pipeline : Standard VDEM pipeline without ray draping
        get_vdem : Core VDEM computation method
        ray_draper : Ray draping geometry handler
        """

        import matplotlib.pyplot as plt
        from matplotlib import colors
        from dateutil.tz import gettz
        from ..tools import (
            xarray_mergeload,
        )
        try:
            import zarr
        except ImportError as err:
            errmsg = ("limb_vdem_pipeline requires zarr; try again after installing, e.g. 'pip install zarr'.")
            raise ImportError(errmsg) from err

        WEST = gettz("US/Western")

        chosen_code = type(self).__name__.replace("Calculator","")
        
        self.snap = snap

        self.units = 'cgs'
        self.emiss_norm = emiss_norm
        self.tile_snaps = False

        coords = self.get_maindims_coords()
        XEXT = coords['x'].max() - coords['x'].min()  # <-- XEXT in un-tiled system, i.e. the original simulation XEXT
        Nx = coords['x'].size
        Ny = coords['y'].size
        Nz = coords['z'].size
        Rsun = DEFAULTS.PHYSICAL.CONSTANTS_SI['R_sun']*1e2  # [cm]
        AU = DEFAULTS.PHYSICAL.CONSTANTS_SI['AU']*1e2  # [cm]

        self.ray_draping = dict(R0=Rsun, T=AU, ray_s=Nz, Asph_theta=Asph_theta, Asph_phi=Asph_phi,
                                fov=(fov_method, 10, 10), Abox_x=int(Abox_x_NXEXT*XEXT))
        ax_sph, ax_box = self.ray_draper.plot('both', periodic=False, fov=False, aspect=(1, 1, 1))
        self.ray_draper.plot_post(-XEXT, 0, ax_sph, ax_box, color='cyan')
        self.ray_draper.plot_post(0,10,ax_sph, ax_box,color="purple")
        # Save the figure
        saveto = f'ray_draper__{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png'
        os.makedirs(dst, exist_ok=True)
        plt.savefig(os.path.join(dst, saveto), dpi=300, bbox_inches='tight')


        self.ray_draping = dict(R0=Rsun, T=AU, ray_s=int(Nz*Nz_times), Asph_theta=Asph_theta, Asph_phi=Asph_phi,
                                fov=(fov_method, int(Nx*Nx_times), int(Ny*Ny_times)), Abox_x=int(Abox_x_NXEXT*XEXT))
        self.ray_draping['fov'] = self.ray_draper.fov

        self.tile_snaps=('x', tile_snaps)
        self.los = 'ray_draping'

        # Checking limits
        tg = self("T")
        ulos = self("u", units = 'si') / 1e3 # to km/s

        # The histogram allows to find the temperature and velocity limits
        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        hist_values = np.log10(tg).plot.hist(density=True, bins=tg_bins, ax=ax[0])

        maxtg = (
                np.ceil(
                    np.max(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * tg_percent / 100.0)])
                    / dlogT
                )
                * dlogT
            )
        mintg = (
                np.floor(
                    np.min(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * tg_percent / 100.0)])
                    / dlogT
                )
                * dlogT
            )
        if mintg_cut is not None:
            mintg = np.max([mintg, mintg_cut])
        if maxtg_cut is not None:
            maxtg = np.min([maxtg, maxtg_cut])

        hist_values = ulos.plot.hist(density=True, bins=ulos_bin, ax=ax[1])
        fig.tight_layout()
        saveto = f"T_ulos_hist_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        fig.savefig(os.path.join(dst, saveto))
        
        maxvel = (
                np.ceil(
                    np.max(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * vel_percent / 100.0)])
                    / dvdop
                )
                * dvdop
            )
        minvel = (
                np.floor(
                    np.min(hist_values[1][np.where(hist_values[0] > np.max(hist_values[0]) * vel_percent / 100.0)])
                    / dvdop
                )
                * dvdop
            )
        if minvel_cut is not None:
            minvel = np.max([minvel, minvel_cut])
        if maxvel_cut is not None:
            maxvel = np.min([maxvel, maxvel_cut])
        
        self.rcoords_logT = np.arange(mintg, maxtg, dlogT)
        self.rcoords_vdop_kms = np.arange(minvel, maxvel, dvdop)

        slicers = [slice(i, None, nslicer) for i in range(nslicer)]

        # 
        self.ray_draping['fov'] = self.ray_draper.fov
        FOV = self.ray_draper.FOV
        for slicer in slicers:
            self.ray_draping['fov'] = FOV.isel(FOVx=slicer)
            vdem = self("vdem")
            if vdem.name is None:
                vdem.name = "vdem"
            vdem.pc.save(f'_pc_temp_fov_lib_vdem/any_filename_you_want__{slicer}')

        vdem = xarray_mergeload('_pc_temp_fov_lib_vdem', kw_merge=dict(join='outer', compat="no_conflicts"))

        if os.path.exists("_pc_temp_fov_lib_vdem"):
            shutil.rmtree("_pc_temp_fov_lib_vdem")

        if self.vdem_varies_logD:
            vdem_sumd = vdem.sum(dim='logD')
        else: 
            vdem_sumd = vdem
        mom = vdem_sumd.pc.moments(dim='vdop')
        fig, ax = plt.subplots(1, 3, figsize=(12, 6))
        mom.moment_0.sum(dim='logT').plot.imshow(norm=colors.LogNorm(vmin=1e-5,vmax=1e5), ax=ax[0])
        mom.moment_1.sum(dim='logT').plot.imshow(ax=ax[1], cmap="bwr")
        mom.moment_2.sum(dim='logT').plot.imshow(ax=ax[2])
        fig.tight_layout()
        saveto = f"vdem_moment_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        fig.savefig(os.path.join(dst, saveto))

        if iz0 is None:
            iz0 = np.argmin(np.abs(self.get_maindims_coords()['z']))
        sl = slice(0,iz0-1)
        self.slices = ({'z':sl})

        nslicer_cut = int(nslicer/3)+1
        slicers = [slice(i, None, nslicer_cut) for i in range(nslicer_cut)]

        coords = self.get_maindims_coords()
        XEXT = coords['x'].max() - coords['x'].min()  # <-- XEXT in un-tiled system, i.e. the original simulation XEXT
        Nx = coords['x'].size
        Ny = coords['y'].size
        Nz = coords['z'].size

        self.ray_draping = dict(R0=Rsun, T=AU, ray_s=int(Nz*Nz_times), Asph_theta=0, Asph_phi=0,
                                fov=(fov_method, int(Nx*Nx_times), int(Ny*Ny_times)), Abox_x=int(Abox_x_NXEXT*XEXT))
        FOV_cut = self.ray_draper.FOV

        for slicer in slicers:
            self.ray_draping['fov'] = FOV_cut.isel(FOVx=slicer)
            vdem_cut = self("vdem", vdem_logT = (0, np.arange(3.,4.4,0.1)), 
                            rcoords_vdop_kms=np.arange(-1.0e1,1.e1,1), 
                            vdem_mode='nointerp', 
                            emiss_mode = 'notrac_noopa',
                            vdem_ignores_photosphere=False, 
                            vdem_loopdim=None)
            if vdem_cut.name is None:
                vdem_cut.name = "vdem"
            vdem_cut.pc.save(f'_pc_temp_fov_lib_vdem_ph/any_filename_you_want__{slicer}')

        vdem_cut = xarray_mergeload('_pc_temp_fov_lib_vdem_ph', kw_merge=dict(join='outer', compat="no_conflicts"))

        if os.path.exists("_pc_temp_fov_lib_vdem_ph"):
            shutil.rmtree("_pc_temp_fov_lib_vdem_ph")

        mom = vdem_cut.pc.moments(dim='vdop')
        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        mom.moment_0.sum(dim='logT').plot(ax=ax[0])
        mom.moment_1.sum(dim='logT').plot(vmin=-5,vmax=5,cmap='bwr', ax=ax[1])
        ax[1].set_title("Doppler velocity (positive must be downward)")
        fig.tight_layout()
        saveto = f"vdem_moment_ph_{chosen_code}{modelname}_{author}_{datetime.datetime.now(WEST).date()!s}_{self.snap}.png"
        fig.savefig(os.path.join(dst, saveto))

        self.ray_draping['fov'] = FOV

        xarray_update_call_history(vdem, locals(), self.limb_vdem_pipeline)

        vdem0 = vdem.to_dataset(name='vdem')
        # _xarray_save_prep(vdem)[0] trashes PlasmaCalcs rich metadata like Snap and Component objects,
        #  and adds some helpful attrs like pc__version__, datetime, and maybe pc__commit_hash.
        vdem0 = _xarray_save_prep(vdem0, add_code_snapshot_info=dict(default='_undefined_'))[0]
        vdem0 = vdem0.rename({"FOVx": "x", "FOVy": "y"})
        vdem0.x.attrs["long_name"] = "X"
        vdem0.y.attrs["long_name"] = "Y"
        vdem0.x.attrs["units"] = "cm"
        vdem0.y.attrs["units"] = "cm"
        vdem0.vdem.attrs["units"] = "1e27 / cm5"
        vdem0.vdem.attrs["description"] = "DEM(T,vel,x,y)"
        vdem0.vdop.attrs["long_name"] = r"v$_{Doppler}$"
        vdem0.vdop.attrs["units"] = "km/s"
        vdem0.logT.attrs["long_name"] = r"log$_{10}$(T)"
        vdem0.logT.attrs["units"] = r"log$_{10}$ (K)"
        vdem0.attrs["emiss_mode"] = self.emiss_mode
        vdem0.attrs["emiss_norm"] = self.emiss_norm
        vdem0.attrs["vdem_mode"] = self.vdem_mode
        vdem0.attrs["los"] = self.los
        vdem0.attrs["tile_snaps"] = tile_snaps
        vdem0.attrs["wavelength"] = self.rcoords_wavelength_A
        now = datetime.datetime.now(WEST).date()
        saveto = f"vdem_{chosen_code}{modelname}_limb_{author}_{now!s}_snaps_{'_'.join(map(str, tile_snaps))}.zarr"
        vdem0.to_zarr(os.path.join(dst, saveto), zarr_version=zarr_version)
        return vdem0



"""
File Purpose: Registry and Registerable class
"""
import collections
import uuid
import re

from .metadata_haver import MetadataHaver, UuidHaver
from .registry_tools import (
    registry_location, registry_schema_sql, registry_name_default,
)
from ..errors import (
    InputError, InputMissingError, InputConflictError,
    DimensionalityError,
    RegistryConflictError, RegistryContentsError,
)
from ..tools import (
    UNSET, simple_property, using_attrs,
    datetime_now, _PlasmaCalcs_version,
    SqlBase,
)
from ..defaults import DEFAULTS


### --------------------- Registry --------------------- ###

class Registry(SqlBase):
    '''registry class which handles adding entries to a registry database (using SQL).
    (Eventually may have engines besides sqlite3, but only using sqlite3 for now.)

    location: UNSET, None, or str
        UNSET --> use registry implied by DEFAULTS.REGISTRY_LOCATION
            (if DEFAULTS.REGISTRY_LOCATION=None, use registry_location_default(),
             which is system-dependent, e.g. "~/.plasmacalcs/pc_registry.db" on Linux.)
        None --> in-memory registry, equivalent to ":memory:", will not edit any files.
        str --> filepath to registry database file.
    name: UNSET, None, or str
        UNSET --> use basename of location. E.g. "pc_registry.db" by default.
            (if location in-memory, equivalent to name=None)
        None --> no name; use '_unnamed_<short_uuid>' as name,
            filling <short_uuid> with the first 8 chars of this registry's source_uuid.
        str --> use this for registry name. Crash if registry already has non-default stored name.
        Note that registry name is stored in registry: with self.connect() as con:
            name == con.execute('SELECT name FROM sources WHERE source_id = 0').fetchone()[0]
    username: UNSET, None, or str
        UNSET --> use DEFAULTS.REGISTRY_USERNAME, which is None by default.
        None --> no known username for the current user. This is fine for local databases.
        str --> username of the current user. Required when working with online databases.
        Expected to be immutable within this Registry object;
            changing self.username after initialization leads to unspecified behavior.
    assert_exists: None, True, or False
        True --> assert that implied location already exists as a file.
        False --> assert that implied location does not yet exist as a file.
        None --> no assertion is made.
        (Here, ':memory:' never "already exists", since it always means a new database.)

    Database schema looks roughly like this:
        users: table of users (e.g. "who created this entry originally")
            user_id (PK): unique id for each user in this database.
                in local database, user_id=0 is reserved for unspecified user, username="__self__".
                in online database, user_id=0 is not allowed and all usernames must be specified.
            username: uniquely identifies this user across all databases.
        sources: table of sources (e.g. "which database did this entry come from")
            source_id (PK): unique id for each source in this database.
                source_id=0 is reserved for "this database".
            name: source name. Duplicates across registries are allowed but discouraged.
                default sql name is "__here__"; python updates to "_unnamed_{start_of_uuid}".
            source_uuid: uniquely identifies this source's database.
                default sql value is a bunch of 0s; python updates to an actual random uuid.
            creator_user_id (FK): user_id of the user who first created this source's database.
        runs: table of runs (or observations or other data source)
            run_id (PK): unique id for each run in this database.
            user_id (FK): tells which user last registered this run (see users table above)
            source_id (FK): tells which source this run came from (see below)
            source_run_id: tells run_id of this entry in its original registry (NULL when source_id=0)
            pc_uuid: unique uuid for this run. Saved to _pc_uuid.txt in the run directory.
            registered_at_first: datetime when this run was first registered
            registered_at: datetime when this run's data was last updated
            calculator_type: name of PlasmaCalculator class which called .register()
            ... more metadata about the run, e.g. datetime_run
        params: table of values of all input parameters across all runs
            param_id (PK): unique id for each param entry
            run_id (FK): run_id associated with this entry
            name: name of this param
            value: value of this param (if numerical), must be a single scalar
            value_text: value of this param (if string or anything more complicated like a list)
        metrics: table of values of all metrics across all runs
            metric_id (PK): unique id for each metric entry
            run_id (FK): run_id associated with this entry
            name: name of this metric
            value: value of this metric (numerical)
            is_pc_quantity: bool telling whether this metric was computed directly as a quantity or not
        register_history: table tracking calls to register() which caused any changes to database.
            history_id (PK): unique id for each history entry
            run_id (FK): run_id associated with this entry
            user_id (FK): user_id telling which user made this register() call.
            source_id (FK): source_id telling which source this register() call occurred on.
            entered_at: datetime when this call to register() occurred
            pc_version: PlasmaCalcs.__version__ when .register() was called
            updated_metadata: list of metadata updated (i.e. value changed) by this call to register()
            updated_metrics: list of metrics updated (i.e. value changed) by this call to register()
    
    Merging databases will alter run_id for runs, such that run_id throughout the result is unique.
    Merged databases will include multiple entries in sources table; unmerged will have only source_id=0.

    Online databases will alter run_id and source_id as needed.
    '''
    RegistrationResult = collections.namedtuple('RegistrationResult', ('run_id', 'history_id'))
    registration_result_cls = RegistrationResult   # override registration_result_cls to utilize different output.

    def __init__(self, location=UNSET, *, name=UNSET, username=UNSET, assert_exists=None):
        '''initialize registry with this location; see help(type(self)) for details.'''
        self.location = registry_location(location, assert_exists=assert_exists)
        self.name = name
        if username is UNSET:
            username = DEFAULTS.REGISTRY_USERNAME
        if (username is None) and (not self.is_local):
            raise InputMissingError('username must be provided for non-local database')
        self.username = username

    # # # SQL CONNECTION AND CROSS-ENGINE COMPATIBILITY # # #
    def schema_exists(self, *, _con=None):
        '''return whether self.location points to a database with registry schema.
        Here, just checks whether "runs", "params", and "metrics" tables exist.
        (This method never initializes schema, regardless of self.connect_can_init)

        _con: None or open connection object
            for efficiency, can provide connection object if already created.
        '''
        if not self._location_exists():
            return False
        with using_attrs(self, connect_can_init=False):
            tables = self.get_tables(_con=_con)
        return all(table in tables for table in ('runs', 'params', 'metrics'))

    connect_can_init = simple_property('_connect_can_init', default=True,
        doc='''whether self.connect() is allowed to initialize things like schema, source name, etc.
        If can init, still only initializes if appropriate, e.g. checks self.schema_exists().
        Default True, i.e. self.connect() initializes schema if needed.''')

    connect_init_only_once = simple_property('_connect_init_only_once', default=True,
        doc='''whether to set self.connect_can_init=False after the first initialization attempt,
        i.e. after the first self.connect() call where self.connect_can_init was True.''')

    def connect(self):
        '''return a connection to this registry's database.

        if self.connect_can_init, also runs some initialization routines as needed:
            - if the database doesn't exist yet (and not online database), create it.
            - if the schema hasn't been initialized yet, initialize it. See: registry_schema_sql()
            - if the database name (sources source_id=0 name) is '__self__', update it to self.name,
                otherwise update self.name to match it.

        Can be used as a context manager to ensure it is closed properly:
            with registry.connect() as con:
                results = con.execute('SELECT ...').fetchall()
        Else, caller is responsible for closing the connection via con.close().
        '''
        con = super().connect()
        with con:
            if self.connect_can_init:
                if not self.schema_exists(_con=con):
                    schema_sql = registry_schema_sql(engine=self.engine_backend)
                    con.executescript(schema_sql)
                self._init_user(con)
                self._init_source(con)
                if self.connect_init_only_once:
                    self.connect_can_init = False
        return con

    def _init_user(self, con):
        '''initialize self.user_id, and users table as appropriate.
        con must be an open connection.
        '''
        if self.username is None:
            self.user_id = 0
        else:
            row = con.execute('SELECT user_id FROM users WHERE username = ?', (self.username,)).fetchone()
            if row is None:
                self.user_id = con.execute('INSERT INTO users (username) VALUES (?) RETURNING user_id',
                                           (self.username,)).fetchone()[0]
            else:
                self.user_id = row[0]

    def _init_source(self, con):
        '''initialize self.name, sources.name, and sources.source_uuid as appropriate.
        con must be an open connection.
        '''
        sourcerow = con.execute('SELECT name, source_uuid FROM sources WHERE source_id = 0').fetchone()
        name, uuid_ = sourcerow
        if (name == '__here__') != (uuid_ == '0'*32):
            errmsg = ('Expected sources.name="__here__" if and only if sources.source_uuid="0"*32.\n'
                      f'Got sources.name={name!r}, source_uuid={uuid!r}.')
            raise RegistryContentsError(errmsg)
        if name == '__here__':
            # new, uninitialized database. Initialize name and source_uuid.
            new_uuid = uuid.uuid4().hex
            name = registry_name_default(self.location, self.name, uuid_=new_uuid)
            con.execute('UPDATE sources SET name = ?, source_uuid = ? WHERE source_id = 0',
                        (name, new_uuid))
        elif isinstance(self.name, str):  # non-default name provided here
            if re.fullmatch(r'_unnamed_[0-9a-f]{8}', name):
                # working with pre-initialized database that has a default name; update it.
                con.execute('UPDATE sources SET name = ? WHERE source_id = 0', (self.name,))
            elif name != self.name:
                errmsg = (f'Provided name={self.name!r} does not match database name={name!r}.\n'
                          '(Database name determined by: SELECT name FROM sources WHERE source_id=0)')
                raise RegistryConflictError(errmsg)
        self.name = name
        # [TODO] should self.name instead update dynamically if a different connection updates sources.name?

    user_id = simple_property('_user_id',
        doc='''id of user associated with this Registry object, based on self.username.
        0 corresponds to unspecified user (only allowed for local databases).
        Set only once, during initialization.''')
    @user_id.getter
    def user_id(self):
        '''returns self._user_id if it exists, else crash with helpful error.'''
        if hasattr(self, '_user_id'):
            return self._user_id
        else:
            errmsg = ('requested self.user_id but it currently does not exist!\n'
                      'user_id is normally set during first call to self.connect().\n'
                      'This crash probably means things were not intialized properly.')
            raise RegistryContentsError(errmsg)

    # # # UNIQUENESS # # #
    def run_id(self, metadata, sanitized_params):
        '''return the run_id of this run within the registry, or None if it doesn't exist yet.
        Compares metadata['pc_uuid'] with [pc_uuid from runs table] to find existing run,
        then compares params (to properly handle the unlikely case of accidental uuid overlap).

        (Raise RegistryContentsError if pc_uuid exists but params don't match.)
        '''
        uuid = metadata['pc_uuid']
        with self.connect() as con:
            row = con.execute('''
                SELECT run_id FROM runs WHERE pc_uuid = ?
            ''', (uuid,)).fetchone()  # schema guarantees unique within db.
            if row is None:
                return None
            run_id = row[0]
        # get params for this run_id; check match below.
        # (conceptually different than matching uuid, but included in same connect() for efficiency.
            fetched_param_rows = con.execute('''
                SELECT name, value, value_text FROM params WHERE run_id = ?
            ''', (run_id,)).fetchall()
        self._assert_all_params_match(sanitized_params, fetched_param_rows, run_id=run_id, uuid=uuid)
        return run_id

    def _assert_all_params_match(self, sanitized_params, fetched_param_rows, *, run_id=None, uuid=None):
        '''check that all sanitized_params match fetched_param_rows from database.
        Crash with RegistryConflictError if params don't all match.
        run_id and uuid go into error message if crash occurs.
        '''
        registered_params = {name: value_text if value is None else value
                                    for name, value, value_text in fetched_param_rows}
        old_keys = set(registered_params.keys())
        new_keys = set(sanitized_params.keys())
        if not new_keys == old_keys:
            extra_new = new_keys - old_keys
            extra_old = old_keys - new_keys
            errmsg = f"Existing run (run_id={run_id}) has same uuid ({uuid!r}) but different params."
            if extra_new:
                errmsg += f"\nNot-yet-registered run has params not in existing run: {extra_new}"
            if extra_old:
                errmsg += f"\nExisting run has params not in not-yet-registered run: {extra_old}"
            raise RegistryConflictError(errmsg)
        # else:  # keys match, need to check values
        for k in old_keys:
            v_old = registered_params[k]
            v_new = sanitized_params[k]
            if v_old != v_new:
                errmsg = (f"Existing run (run_id={run_id}) has same uuid ({uuid!r}) but different params."
                          f"Param {k!r}: existing_value={v_old!r}; not_yet_registered_value={v_new!r}")
                raise RegistryConflictError(errmsg)
        return True  # not necessary but explicit


    _REGISTER_META_METADATA = ('user_id', 'source_id', 'source_run_id', 'registered_at_first', 'registered_at')

    def run_metadata_and_sanitized_params(self, run_id):
        '''returns dict of metadata and dict of sanitized_params, for run with this run_id in self.
        metadata includes all "run" metadata but not any of the "register-related" metadata,
            i.e. it excludes source_id, source_run_id, registered_at_first, registered_at.
        '''
        with self.connect() as con:
            metadata_row = con.execute('''
                SELECT * FROM runs WHERE run_id = ?
            ''', (run_id,)).fetchone()
            if metadata_row is None:
                raise RegistryContentsError(f"run_id {run_id} not found in registry.")
            metadata_columns = self.columns('runs')
            metadata = {col: val for col, val in zip(metadata_columns, metadata_row)
                        if col not in self._REGISTER_META_METADATA}
            param_rows = con.execute('''
                SELECT name, value, value_text FROM params WHERE run_id = ?
            ''', (run_id,)).fetchall()
            sanitized_params = {name: value_text if value is None else value
                                    for name, value, value_text in param_rows}
        return metadata, sanitized_params

    # # # REGISTER # # #
    def register(self, metadata, params, metrics):
        '''register an entry with this metadata, params, and metrics.
        Adds some registration-related metadata like registered_at = current datetime.
        Returns (run_id, history_id) of this run and register_history entry within the registry.
            (history_id will be None if this run already existed and no values were updated.)
            (output will be a namedtuple to clarify meaning; can be treated as a tuple if desired.)

        metadata: dict
            metadata about the data source (simulation run / observation / other data source)
            Will be stored in the "runs" table.
        params: dict
            parameters used to generate the data source
            Will be sanitized then stored in the "params" table.
            (Sanitizing converts to numerical value or string representation.)
        metrics: dict of (bool, value)
            metrics computed from the data source
            Will be stored in the "metrics" table.
            bool indicates whether this metric was computed directly as a quantity.
                If True, it was computed via calculator(key).
                Else, it was provided directly by user.
        '''
        sanitized_params = self.sanitize_values(params)
        run_id = self.run_id(metadata, sanitized_params)
        kws = dict(run_id=run_id, metadata=metadata, sanitized_params=sanitized_params, metrics=metrics)
        if run_id is None:  # insert new run.
            (run_id, history_id) = self._register_new_run(**kws)
        else:
            (run_id, history_id) = self._register_update_run(**kws)
        return self.registration_result_cls(run_id, history_id)

    def _register_new_run(self, *, run_id=None, metadata, sanitized_params, metrics):
        '''registers new run, returning (run_id, history_id) of the newly registered run.
        Use register() instead; not for direct usage (but subclass can override). 
        Assumes, without checking:
            - self.run_id(metadata, sanitized_params) is None
            - sanitized_params is a dict with values only numbers and strings.

        run_id is provided by self.register() in case subclass wants to use it,
            but the implementation here does not check provided run_id at all.
        '''
        with self.connect() as con:
            now = datetime_now(tz='utc', sep='T')
            # insert into runs table (adding metadata columns as needed first)
            missing = self.missing_columns('runs', metadata.keys())
            for col in missing:
                con.execute(f'ALTER TABLE runs ADD COLUMN "{col}" TEXT')
            # # convert to str to ensure consistency with str conversion in _register_update_run()
            metadata_vals = {**metadata}
            text_metadata_cols = self.text_columns('runs')
            for k in set(text_metadata_cols).intersection(metadata.keys()):
                metadata_vals[k] = str(metadata_vals[k])
            values = {'source_id': 0,  # all runs registered directly get source_id = 0
                      'user_id': self.user_id,
                      **metadata_vals,
                      'registered_at_first': now, 'registered_at': now}
            run_id = con.execute(f'''
                INSERT INTO runs ({', '.join([f'"{k}"' for k in values])})
                VALUES ({', '.join([f':{k}' for k in values])})
                RETURNING run_id
            ''', values).fetchone()[0]
            # insert into params table
            for name, value in sanitized_params.items():
                if isinstance(value, str):
                    con.execute('''
                        INSERT INTO params (run_id, name, value_text)
                        VALUES (?, ?, ?)
                    ''', (run_id, name, value))
                else:
                    con.execute('''
                        INSERT INTO params (run_id, name, value)
                        VALUES (?, ?, ?)
                    ''', (run_id, name, value))
            # insert into metrics table
            for name, (is_pc_quantity, value) in metrics.items():
                con.execute('''
                    INSERT INTO metrics (run_id, name, value, is_pc_quantity)
                    VALUES (?, ?, ?, ?)
                ''', (run_id, name, value, is_pc_quantity))
            # insert into register_history table
            history = {
                'source_id': 0,
                'run_id': run_id,
                'user_id': self.user_id,
                'entered_at': now,
                'pc_version': _PlasmaCalcs_version(),
                'updated_metadata': ', '.join(metadata.keys()),  # all were "updated"; this is a new run.
                'updated_metrics': ', '.join(metrics.keys()),  # all were "updated"; this is a new run.
            }
            history_id = con.execute('''
                INSERT INTO register_history (source_id, run_id, user_id, entered_at, pc_version,
                                              updated_metadata, updated_metrics)
                VALUES (:source_id, :run_id, :user_id, :entered_at, :pc_version,
                        :updated_metadata, :updated_metrics)
                RETURNING history_id
            ''', history).fetchone()[0]
        return (run_id, history_id)

    def _register_update_run(self, *, run_id, metadata, sanitized_params=None, metrics):
        '''updates metadata, metrics, and register_history for existing run with this run_id.
        Use register() instead; not for direct usage (but subclass can override).
        Assumes, without checking:
            - run_id exists in database and has all matching sanitized_params for this run

        sanitized_params is provided by self.register() in case subclass wants to use it,
            but the implementation here does not check provided sanitized_params at all.

        returns (run_id, history_id) of the updated run and register_history entry.
            (history_id will be None if no values were updated.)
        '''
        with self.connect() as con:
            now = datetime_now(tz='utc', sep='T')
            # update user_id
            user_id = con.execute('''
                UPDATE runs SET user_id = :suid WHERE run_id = :run_id AND user_id IS DISTINCT FROM :suid
                RETURNING user_id
            ''', dict(suid=self.user_id, run_id=run_id)).fetchone()
            updated_user_id = user_id is not None
            # update metadata table
            updated_metadata = []
            missing_metakeys = self.missing_columns('runs', metadata.keys())
            text_metadata_cols = self.text_columns('runs')
            for key, value in metadata.items():
                if key in text_metadata_cols or key in missing_metakeys:
                    value = str(value)   # <-- convert to string for TEXT type.
                if key in missing_metakeys:  # key is missing --> add column
                    con.execute(f'ALTER TABLE runs ADD COLUMN "{key}" TEXT')
                    updated = True
                else:  # key exists --> update value if it changed
                    _run_id = con.execute(f'''
                        UPDATE runs
                        SET "{key}" = :value
                        WHERE run_id = :run_id
                            AND "{key}" IS DISTINCT FROM :value
                        RETURNING run_id   -- returns something if value actually changed
                    ''', dict(value=value, run_id=run_id)).fetchone()
                    updated = (_run_id is not None)
                # if updated, track it for register_history:
                if updated:
                    updated_metadata.append(key)
            # update metrics table:
            updated_metrics = []
            for name, (is_pc_quantity, value) in metrics.items():
                metric_info = {
                    'run_id': run_id,
                    'name': name,
                    'value': value,
                    'is_pc_quantity': is_pc_quantity,
                }
                # if (run_id, name) is missing, insert it, else update value (if it changed)
                metric_exists = con.execute('''
                    SELECT metric_id from metrics
                    WHERE run_id = :run_id AND name = :name
                    ''', metric_info).fetchone()
                metric_missing = (metric_exists is None)
                if metric_missing:  # (run_id, name) is missing --> insert
                    con.execute('''
                        INSERT INTO metrics (run_id, name, value, is_pc_quantity)
                        VALUES (:run_id, :name, :value, :is_pc_quantity)
                    ''', metric_info)
                    updated = True
                else:  # (run_id, name) exists --> update (value, is_pc_quantity) if it changed
                    _metric_id = con.execute('''
                        UPDATE metrics
                        SET value = :value, is_pc_quantity = :is_pc_quantity
                        WHERE run_id = :run_id AND name = :name
                            AND (   -- check if value actually changed.
                                 value IS DISTINCT FROM :value
                                 OR is_pc_quantity IS DISTINCT FROM :is_pc_quantity
                                )
                        RETURNING metric_id   -- returns something if value actually changed
                    ''', metric_info).fetchone()
                    updated = (_metric_id is not None)
                # if updated, track it for register_history:
                if updated:
                    updated_metrics.append(name)
            # if anything was actually updated, add register_history entry and update registered_at
            if updated_user_id or updated_metadata or updated_metrics:
                update_info = {
                    'source_id': 0,
                    'run_id': run_id,
                    'user_id': self.user_id,
                    'registered_at': now,
                    'entered_at': now,
                    'pc_version': _PlasmaCalcs_version(),
                    'updated_metadata': ', '.join(updated_metadata),
                    'updated_metrics': ', '.join(updated_metrics),
                }
                # update registered_at
                con.execute('''
                    UPDATE runs
                    SET registered_at = :registered_at
                    WHERE run_id = :run_id
                ''', update_info)
                # update register_history
                history_id = con.execute('''
                    INSERT INTO register_history (source_id, run_id, user_id, entered_at, pc_version,
                                                  updated_metadata, updated_metrics)
                    VALUES (:source_id, :run_id, :user_id, :entered_at, :pc_version,
                            :updated_metadata, :updated_metrics)
                    RETURNING history_id
                ''', update_info).fetchone()[0]
            else:
                history_id = None
        return (run_id, history_id)

    # # # MERGING / HANDLING SOURCE_ID # # #
    def merge(self, *others, dst=None):
        '''merge self other registry(s), writing result to dst, and returning Registry(dst).
        dst must represent a new registry; this operation never alters an existing registry.

        For any runs appearing in multiple registries (based on pc_uuid),
            only keeps values from first registry in which is appears.
            ([TODO] more leniency, e.g. also merge those runs' metrics data from later registries.)

        others: each is a Registry or str
            each is a registry to merge into the result.
            for str, assert it points to an existing file, then use other=type(self)(dst).
        dst: None or str
            destination for merge operation.
            None --> like location=None or ":memory:"; creates a new in-memory registry.
            str --> create registry at this filepath. It must not yet exist.
                (internally, appends ".db" extension if no file extension detected.)
        '''
        # checks & setup:
        for i, other in enumerate(others):
            if isinstance(other, str):
                others[i] = type(self)(other, assert_exists=True)
        result = type(self)(dst, assert_exists=False)

        # adding registries' data to result, one at a time:
        with result.connect() as res:
            for reghere in (self, *others):
                with reghere.connect() as here:
                    # can't just do a full backup, because we need to put new user_id and source_id appropriately.
                    # (we also need to add source_run_id to each run)
                    # so we iterate through runs, copying them one at a time.
                    # first, iterate through users:
                    users_here = here.execute('SELECT user_id, username FROM users').fetchall()
                    map_user_id = {}
                    for user_id, username in users_here:
                        # check if username exists already; use corresponding user_id if it does.
                        new_user_id = res.execute('SELECT user_id FROM users WHERE username = ?',
                                                  (username,)).fetchone()
                        if new_user_id is None:
                            new_user_id = res.execute('''
                                INSERT INTO users (username)
                                VALUES (?)
                                RETURNING user_id
                            ''', (username,)
                            ).fetchone()
                        map_user_id[user_id] = new_user_id[0]
                    # next, iterate through sources:
                    sources_here = here.execute('''
                        SELECT source_id, name, source_uuid, creator_user_id FROM sources
                    ''').fetchall()
                    map_source_id = {}
                    for source_id, source_name, source_uuid, creator_user_id in sources_here:
                        new_source_id = res.execute('SELECT source_id FROM sources WHERE source_uuid = ?',
                                                    (source_uuid,)).fetchone()
                        if new_source_id is None:
                            new_source_id = res.execute('''
                                INSERT INTO sources (name, source_uuid, creator_user_id)
                                VALUES (?, ?, ?)
                                RETURNING source_id
                            ''', (source_name, source_uuid, map_user_id[creator_user_id])
                            ).fetchone()
                        map_source_id[source_id] = new_source_id[0]
                    # next, iterate through runs:  # [TODO] any way to bulk insert runs?
                    map_run_id = {}
                    run_columns = reghere.columns('runs')
                    runs_here = here.execute('SELECT * FROM runs').fetchall()
                    for run in runs_here:
                        run_dict = dict(zip(run_columns, run))
                        old_run_id = run_dict['run_id']
                        run_dict['user_id'] = map_user_id[run_dict['user_id']]
                        run_dict['source_id'] = map_source_id[run_dict['source_id']]
                        run_dict['source_run_id'] = old_run_id
                        del run_dict['run_id']
                        metadata, sanitized_params = reghere.run_metadata_and_sanitized_params(old_run_id)
                        new_run_id = result.run_id(metadata, sanitized_params)
                        if new_run_id is None:  # i.e. run doesn't exist in result yet
                            for col in result.missing_columns('runs', run_columns):
                                res.execute(f'ALTER TABLE runs ADD COLUMN "{col}" TEXT')
                            new_run_id = res.execute(f'''
                                INSERT INTO runs ({', '.join([f'"{k}"' for k in run_dict])})
                                VALUES ({', '.join([f':{k}' for k in run_dict])})
                                RETURNING run_id
                            ''', run_dict).fetchone()[0]
                            map_run_id[old_run_id] = new_run_id
                        # else, don't add to map_run_id; not going to alter result's existing values for this run.
                    # add all params, metrics, and register_history entries.
                    for table in ('params', 'metrics', 'register_history'):
                        pk = reghere.primary_key(table)
                        columns = [k for k in reghere.columns(table) if k != pk]
                        here.execute('CREATE TEMPORARY TABLE _rids (v INTEGER)')  # run_ids we need to merge into res
                        here.executemany('INSERT INTO _rids (v) VALUES (?)',
                                         [(run_id,) for run_id in map_run_id.keys()])
                        rows = here.execute(f'''
                            SELECT {', '.join(columns)} FROM {table}
                            JOIN _rids ON {table}.run_id = _rids.v
                        ''').fetchall()
                        here.execute('DROP TABLE _rids')
                        rows = [list(row) for row in rows]  # make mutable for updating run_id and source_id
                        i_run_id = columns.index('run_id')
                        for row in rows:
                            row[i_run_id] = map_run_id[row[i_run_id]]  # update run_id to new run_id
                        if 'user_id' in columns:
                            i_user_id = columns.index('user_id')
                            for row in rows:
                                row[i_user_id] = map_user_id[row[i_user_id]]
                        if 'source_id' in columns:
                            i_source_id = columns.index('source_id')
                            for row in rows:
                                row[i_source_id] = map_source_id[row[i_source_id]]
                        res.executemany(f'''
                            INSERT INTO {table} ({', '.join(columns)})
                            VALUES ({', '.join(['?'] * len(columns))})
                        ''', rows)
        return result

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = [f'name={self.name!r}']
        result.extend(super()._repr_contents())
        return result


### --------------------- Registerable --------------------- ###

class Registerable(UuidHaver, MetadataHaver):
    '''class whose instances can be registered to a registry, an sql database.'''

    registry_cls = Registry  # class used to create registry from provided "registry" input

    # # # SUBCLASS SHOULD IMPLEMENT # # #
    def registry_params(self):
        '''return flat dict of all input params to use during self.register()
        [Not yet implemented; subclass should implement]
        '''
        raise NotImplementedError(f'{type(self).__name__}.registry_params() required by register().')

    # # # REGISTER AND HELPER METHODS # # #
    def register(self, registry=UNSET, *, metrics=[], username=UNSET):
        '''register this instance and these metrics to the registry.
        returns (registry.register() result, Registry object).
            (registry.register() result is a namedtuple of (run_id, history_id))

        registry: UNSET, None, Registry, or str
            Registry --> use it directly
            else --> pass to self.registry_cls (probably Registry), which probably does:
                UNSET --> use registry implied by DEFAULTS.REGISTRY_LOCATION
                    (if DEFAULTS.REGISTRY_LOCATION=None, use registry_location_default(),
                     which is system-dependent, e.g. "~/.plasmacalcs/pc_registry.db" on Linux.)
                None --> use in-memory registry, equivalent to ":memory:", does not edit any files.
                str --> filepath to registry database file.
        metrics: list of (str or dict) entries
            metrics to register (and possibly compute) for this instance
            str --> compute this metric via self(metric)
            dict --> {key: value} to register directly, without computing via self(key)
            In case of conflict, later values override earlier ones,
                e.g. [{'mykey': 1}, {'mykey': 2}] registers 'mykey' with value 2.
        username: UNSET, None, or str
            Username associated with this operation.
            UNSET --> get value from DEFAULTS.REGISTRY_USERNAME (default None).
            None --> no username provided. Only allowed for local databases.
            str --> username associated with this operation.
                    Also, checks for consistency with registry.username if `registry` is a Registry object.
        '''
        if isinstance(username, str) and isinstance(registry, Registry):
            if registry.username != username:
                errmsg = f"registry.username={registry.username!r} disagrees with provided username={username!r}"
                raise InputConflictError(errmsg)
                # [TODO] if registry.username is None, temporarily update it to the username provided here?
        if not isinstance(registry, Registry):
            registry = self.registry_cls(registry, username=username)
        metadata = self._registry_metadata()
        params = self.registry_params()
        metrics = self._registry_metrics(metrics)
        return (registry.register(metadata, params, metrics), registry)

    def _registry_metadata(self):
        '''return metadata relevant to registering this run.
        This includes all of self.metadata, along with calculator_type and pc_uuid.
        (Uses self.metadata not self.metadata_all; excludes any missing or ATTR_UNSET.)
        If self.pc_uuid does not exist, it will be created via self.get_pc_uuid().
        '''
        metadata = {**self.metadata}  # already excludes ATTR_UNSET
        metadata['calculator_type'] = type(self).__name__
        if 'pc_uuid' not in metadata:
            # generate if needed (or crashes with clear error message if not implemented).
            metadata['pc_uuid'] = self.get_pc_uuid()
        return metadata

    def _registry_metrics(self, metrics):
        '''return dict of metrics' values relevant to registering this run with these metrics.
        Strings in metrics will be converted to {key: self(key)} entries.
            e.g. metrics=['mean_T'] --> result is {'mean_T': self('mean_T')}
        dicts will be used as-is.

        Result will be a dict of {key: (is_pc_quantity, value)} entries,
            with is_pc_quantity telling whether metric was computed via self(key),
            i.e. whether it was a string in the input metrics list.
        '''
        if any(isinstance(m, str) for m in metrics) and getattr(self, 'units', 'si') != 'si':
            raise NotImplementedError('registering non-SI units not yet supported.')
            # ^ technically the code itself can handle it just fine,
            # but this restriction ensures no accidental units confusion.
            # (This can eventually be removed if we record used behavior_attrs somewhere.)
        result = {}
        for i, item in enumerate(metrics):  # TODO: progress updates if slow
            if isinstance(item, str):
                value = self(item)
                if value.size == 1:
                    value = value.item()
                else:
                    errmsg = f'Metrics must be scalars; {item!r} has shape {value.shape}'
                    raise DimensionalityError(errmsg)
                result[item] = (True, value)
            elif isinstance(item, dict):
                for key, value in item.items():
                    result[key] = (False, value)
            else:
                raise InputError(f'Metrics must be str or dict, got {type(item)} at index={i}.')
        return result

    # [TODO] registry_run_id() which returns run_id of this run within the registry or None.



"""
File Purpose: misc tests for MultiCalculator with EppicCalculator calculators
"""

import os

import numpy as np
import xarray as xr

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


### --------------------- can create multicalculator --------------------- ###

def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec

def test_make_multicalculator():
    '''ensure can make a multicalculator.'''
    ecA = get_eppic_calculator()
    ecB = get_eppic_calculator()
    pc.MultiCalculator({'ecA':ecA, 'ecB':ecB})


### --------------------- can load vars --------------------- ###

def test_multicalculator_load_misc():
    '''ensure can load various values via multicalculator'''
    ecA = get_eppic_calculator()
    ecB = get_eppic_calculator()
    mc = pc.MultiCalculator({'ecA':ecA, 'ecB':ecB})
    # load n
    dos = mc('n')  # DictOfSimilar
    assert np.all(dos['ecA'] == dos['ecB'])
    arr = dos.to_xr('calculator')
    assert 'calculator' in arr.dims
    assert arr.sizes['calculator'] == 2
    assert np.all(arr.isel(calculator=0) == arr.isel(calculator=1))
    # setting snap
    assert ecA.snap != 3
    assert ecB.snap != 3
    mc.snap = 3
    assert ecA.snap == 3
    assert ecB.snap == 3
    # setting fluid
    mc.fluid = pc.ELECTRON
    assert ecA.fluid == pc.ELECTRON
    assert ecB.fluid == pc.ELECTRON
    arr = mc('n').to_xr()
    assert np.all(arr.isel(variable=0) == arr.isel(variable=1))
    # ensure .to_xr() works on Dataset values too
    arr = mc('stats_n', snap=[2,5,7], fluid=pc.IONS).to_xr('calculator')
    assert set(arr.dims) == {'fluid', 'snap', 'calculator'}
    # ensure can do various operations to DictOfSimilar:
    dos = mc('n')
    arr = dos.to_xr()
    assert np.all((dos + 1e7)['ecA'] == dos['ecA'] + 1e7)
    assert np.all((1e7 + dos)['ecB'] == dos['ecB'] + 1e7)
    assert np.all((dos * 2)['ecB'] == dos['ecB'] * 2)
    assert np.all((2 * dos)['ecA'] == dos['ecA'] * 2)
    assert np.all((dos / 5).to_xr() == arr / 5)
    assert np.all((5 / dos).to_xr() == 5 / arr)
    assert np.all((dos - 1e7).to_xr() == arr - 1e7)
    assert np.all((1e7 - dos).to_xr() == 1e7 - arr)
    assert np.all((dos**2).to_xr() == arr**2)
    # other misc DictOfSimilar things:
    assert len(dos.subset(keep=['ecA'])) == 1
    assert len(dos.subset(keep=lambda k,v: k=='ecB')) == 1
    # ensure different results if different snaps
    ecA.snap = [3,6,9]
    ecB.snap = [2,4,6]
    dos = mc('meant_n')
    assert np.any(dos['ecA'] != dos['ecB'])
    # misc operations:
    assert dos.apply(lambda arr: arr.mean()).to_xr().size == 2



"""
File Purpose: EppicInputDeck - values from eppic.i file.
"""
import os

import numpy as np

from .eppic_dimensions import (
    EppicDist, EppicDistList,
    EppicNeutral, EppicNeutralList,
)
from .eppic_io_tools import read_eppic_i_file, infer_eppic_dist_names
from ...units import UnitsManagerPIC
from ...tools import (
    alias,
    DictlikeFromKeysAndGetitem,
)
from ...defaults import DEFAULTS


### --------------------- EppicInputDeck --------------------- ###

class EppicInputDeck(DictlikeFromKeysAndGetitem):
    '''info from eppic.i file, in a class.
    Use EppicInputDeck.from_file() to create an instance from a file.
    units are "raw". Numerical values equal those from the eppic.i file.

    params: dict
        values of params
    distributions: dict or EppicDistList
        dict of {distribution number: EppicDist}, or an already-created EppicDistList.
        if dict, will be internally stored as EppicDistList.from_dict(distributions), instead.
    filename: None or str
        path to eppic.i file this came from. E.g. 'path/to/eppic.i'
        str will be internally stored as os.path.abspath(filename), instead.

    Examples:
        eid = EppicInputDeck.from_file('eppic.i')
        eid[1]          # EppicDist for distribution 1
        eid['nz']       # number of grid points in z direction
        eid.params      # dict of all global params
        eid.dists       # EppicDistList of all the distributions
        eid.dists.get('e')    # get EppicDist with name 'e'
        eid.dists.get(0)      # get EppicDist number 0 from this list
        eid[2]['m']           # mass parameter for distribution 2
        eid[0]['coll_rate']   # collision rate parameter for distribution 0
        eid.keys()      # available keys
        eid.items()     # available (key, value) pairs
        eid[0].keys()   # available keys for distribution 0
        eid.get('k', 7) # eid['k'] if it exists, else 7.
    '''
    dist_type = EppicDist  # type for distributions
    dlist_type = EppicDistList  # type for list of distributions
    neutral_type = EppicNeutral  # type for neutral fluid
    nlist_type = EppicNeutralList  # type for list of neutral fluids

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, params, distributions, *, filename=None, **kw_super):
        self.params = params
        self.distributions = distributions if isinstance(distributions, self.dlist_type) \
                                else self.dlist_type.from_dict(distributions)
        self._init_neutral_fluid()
        self.filename = None if filename is None else os.path.abspath(filename)
        super().__init__(**kw_super)

    dists = alias('distributions')

    @classmethod
    def from_file(cls, filename='eppic.i', *, dist_names=None, **kw__init):
        '''return EppicInputDeck from eppic.i file.
        filename: str
            path to eppic.i file.
            abspath(filename) will be saved to result, for easier bookkeeping.
        dist_names: None or dict
            {N: name for distribution N} for any number of distributions in filename.
            None --> infer from file.
        '''
        params, dists = read_eppic_i_file(filename)
        if dist_names is None:
            dist_names = infer_eppic_dist_names(filename)
        if dist_names is not None:
            for N, name in dist_names.items():
                if name is not None:
                    dists[N][f'name{N}'] = name    # (the {N} in the key is required, then removed, by EppicDist.from_dict)
        dist_dict = cls.dist_type.dists_from_dict(dists)
        return cls(params, dist_dict, filename=filename, **kw__init)

    def _init_neutral_fluid(self):
        '''sets self.neutral and self.neutrals based on params of self.
        Particularly, uses params['m_neutral'].
        '''
        m = self.params['m_neutral']
        vth = self.params.get('vth_neutral', None)
        self.neutral = self.neutral_type(m, vth=vth)
        self.neutrals = self.nlist_type([self.neutral])

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'with {len(self.dists)} distributions']
        if self.filename is not None:
            contents.append(f'filename={self.filename!r}')
        contents_str = ', '.join(contents)
        return f'{type(self).__name__}({contents_str})'

    # # # GETTING VALUES & ITERATING # # #
    def __getitem__(self, key):
        '''return self.params[key] if key in self.params, else self.dists[key].'''
        try:
            return self.params[key]
        except KeyError:
            pass  # handled below, to make clearer error message if failing.
        try:
            return self.dists[key]
        except (KeyError, IndexError, ValueError):
            errmsg = f'key {key!r} not in {type(self).__name__}.params or .dists'
            raise KeyError(errmsg) from None

    def keys(self):
        '''return tuple of keys to access self.dists or self.params'''
        return tuple(key for key in (*range(len(self.dists)), *self.params.keys()))

    # # # FILE PATH # # #
    @property
    def dirname(self):
        '''os.path.dirname(self.filename).
        if self.filename is None, raise FileNotFoundError.
        '''
        filename = self.filename
        if filename is None:
            raise FileNotFoundError('self.filename is None; cannot determine dirname.')
        return os.path.dirname(filename)

    # # # DIMENSIONS # # #
    def maindims(self):
        '''returns tuple of main dimensions in this input deck.
        inferred from self['nx'], 'ny', 'nz', and 'ndim_space'
        '''
        ndim_space = self['ndim_space']
        if ndim_space == 3:
            return ('x', 'y', 'z')
        elif ndim_space == 2:
            nx = self.get('nx', None)
            if nx is None: return ('y', 'z')
            ny = self.get('ny', None)
            if ny is None: return ('x', 'z')
            nz = self.get('nz', None)
            if nz is None: return ('x', 'y')
            # provided nx, ny, and nz...
            if nz == 1: return ('x', 'y')
            elif ny == 1: return ('x', 'z')
            elif nx == 1: return ('y', 'z')
            # nx, ny, and nz were all provided, and not 1.
            if getattr(self, 'verbose', True) and not getattr(self, 'warned_ndim2', False):
                print("warning, ndim_space=2 but input deck contains 'nx', 'ny', and 'nz', all != 1.")
                print("assuming space dims ('x', 'y').")
                self.warned_ndim2 = True  # only print the warning once per input_deck.
            return ('x', 'y')  # by default, return ('x', 'y').
        else:
            raise NotImplementedError(f'{type(self).__name__}.maindims() when ndim_space={ndim_space}')

    def _get_nspace_simple(self, xstr):
        '''get "simple" number of cells in {x} dimension. == n{x} // nout_avg
        xstr should be 'x', 'y', or 'z'.
        Note that for x='x', actual number of cells may be affected by nsubdomains.
        See also: self._get_nspace.
        '''
        return self.get(f'n{xstr}', 1) // self.get('nout_avg', 1)

    def _get_velocity_nspace_simple(self, xstr, N):
        '''get "simple" number of cells in {vx} dimension. == pn{vx}{N}
        xstr should be 'vx', 'vy', or 'vz'.
        N is the distribution number.
        Note that for x='x', actual number of cells may be affected by nsubdomains.
        See also: self.get_nspace.
        '''
        N = str(N)
        return self.get(f'pn{xstr}{N}', 1)

    def get_nspace(self, xstr):
        '''get number of cells in {x} dimension, accounting for nout_avg and nsubdomains.
        See also: Nx, Ny, Nz.
        '''
        result = self._get_nspace_simple(xstr)
        if xstr == 'x':
            result = result * self.get('nsubdomains', 1)
        return result

    Nx = property(lambda self: self.get_nspace('x'),
                  doc='''actual total number of grid cells in the x dimension.''')
    Ny = property(lambda self: self.get_nspace('y'),
                  doc='''actual total number of grid cells in the y dimension.''')
    Nz = property(lambda self: self.get_nspace('z'),
                  doc='''actual total number of grid cells in the z dimension.''')

    def get_velocity_nspace(self, xstr, N):
        '''get number of cells in {vx} dimension.
        N is the distribution number.
        See also: Nvx, Nvy, Nvz.
        '''
        result = self._get_velocity_nspace_simple(xstr, N)
        return result

    Nvx = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vx dimension (ordered by distribution).''')
    Nvy = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vy dimension (ordered by distribution).''')
    Nvz = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vz dimension (ordered by distribution).''')

    def get_dspace(self, xstr):
        '''get d{x} [raw units] between output cells.
        accounts for nout_avg; see also: Dx, Dy, Dz.
        if {x} not provided in self (e.g. 2D in x & y, xstr='z'), returns 1 * nout_avg.
        '''
        return self.get(f'd{xstr}', 1) * self.get('nout_avg', 1)

    Dx = property(lambda self: self.get_dspace('x'),
                    doc='''dx between adjacent output cells [raw units]''')
    Dy = property(lambda self: self.get_dspace('y'),
                    doc='''dy between adjacent output cells [raw units]''')
    Dz = property(lambda self: self.get_dspace('z'),
                    doc='''dz between adjacent output cells [raw units]''')

    def _get_velocity_extents(self, xstr, N):
        '''get (p{vx}min{N}, p{vx}max{N}) [raw units].
        N is the distribution number.
        See also: Dvx, Dvy, Dvz.
        '''
        return (self.get(f'p{xstr}min{N}', 1), self.get(f'p{xstr}max{N}', 1))

    def get_coord(self, xstr):
        '''get coordinate array [raw units] for this dimension.
        xstr: str
            dimension name. 'x', 'y', or 'z'.
        '''
        dx = self.get_dspace(xstr)
        Nx = self.get_nspace(xstr)
        return np.arange(Nx) * dx

    def get_space_coords(self):
        '''return dict of {x: xcoord array} for x in self.maindims(), a subset of ('x', 'y', 'z').'''
        return {x: self.get_coord(x) for x in self.maindims()}

    def get_dspace_min(self):
        '''return min of self.Dx, self.Dy, self.Dz.'''
        return min(self.Dx, self.Dy, self.Dz)

    def get_dspace_max(self):
        '''return max of self.Dx, self.Dy, self.Dz.'''
        return max(self.Dx, self.Dy, self.Dz)

    # # # UNITS # # #
    def get_units_manager(self, *, u_l=None, u_t=None, u_n=None, ne_si=None, **kw_u_from_pic):
        '''return UnitsManagerPIC for this input deck. Assumes distribution 0 is electrons.
        Must provide u_l, u_t, u_n, or ne_si.
        Will read ne, qe, me, and eps0 from the input deck, unless they are provided in kw_u_from_pic.
        See help(UnitsManagerPIC.pic_ambiguous_unit) for details.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if 'ne' not in kw_u_from_pic: kw_u_from_pic['ne'] = self[0]['n0']
        if 'qe' not in kw_u_from_pic: kw_u_from_pic['qe'] = abs(self[0]['q'])
        if 'me' not in kw_u_from_pic: kw_u_from_pic['me'] = self[0]['m']
        if 'eps0' not in kw_u_from_pic: kw_u_from_pic['eps0'] = self['eps']
        return UnitsManagerPIC.from_pic(u_l=u_l, u_t=u_t, u_n=u_n, ne_si=ne_si, **kw_u_from_pic)

    # # # QUASINEUTRAL? # # #
    def is_quasineutral(self):
        '''return whether this input deck is in quasineutral mode.
        as a proxy, guessing quasineutral iff any dist.is_hybrid() for dist in self.dists.
        '''
        return any(dist.is_hybrid() for dist in self.dists)

    # # # REGISTRY PARAMS # # #
    def registry_params(self):
        '''returns flat dict of all input params relevant to registry for calculator with this input deck.
        
        includes all values from self.params.
        Also all values from self.dists,
            though, values from self.dists have N appended, e.g. 'md' becomes 'md1',
            similarly to eppic.i file (but not necessarily exactly the same as original file;
            eppic itself has some tolerance for params without N appended, but here always has N.
        '''
        result = {**self.params}
        for dist in self.dists:
            N = int(dist)
            # STYLE: if 'name' in dist.params, put name first instead of last:
            if 'name' in dist.params:
                result[f'name{N}'] = dist.params['name']
            # then put all other params:
            for k, v in dist.params.items():
                if k != 'name':
                    result[f'{k}{N}'] = v
        return result



"""
File Purpose: testing MhdRadiativeLoader functionality, using bifrost data as example.
"""
import os

import numpy as np
import pytest

import PlasmaCalcs as pc

HERE = os.path.dirname(__file__)


### --------------------- making bifrost calculator --------------------- ###

def get_bifrost_calculator(**kw_init):
    '''make a bifrost calculator'''
    with pc.InDir(os.path.join(HERE, 'test_bifrost_tinybox')):
        bc = pc.BifrostCalculator(**kw_init)
    return bc

def get_bifrost_multifluid_calculator(**kw_init):
    '''make a multifluid bifrost calculator (for multifluid analysis of bifrost outputs)'''
    with pc.InDir(os.path.join(HERE, 'test_bifrost_tinybox')):
        bm = pc.BifrostMultifluidCalculator(**kw_init)
    return bm


### --------------------- getting various vars --------------------- ###

def test_bifrost_emiss_vars():
    '''testing emiss-related vars and options, using Bifrost data'''
    bc = get_bifrost_calculator()
    bm = get_bifrost_multifluid_calculator()
    # emiss stuff (vdems need emiss)
    bc.tree('emiss')   # just checking it doesn't crash
    bm.tree('emiss')
    bc.emiss_mode = 'notrac_noopa'
    tree1 = bc.tree('emiss')
    assert tree1.contains_var('emiss_notrac')
    assert not tree1.contains_var('emiss_trac')
    assert not tree1.contains_var('tau')
    tree2 = bc.tree('emiss', emiss_mode='trac_opa')
    assert tree2.contains_var('emiss_trac')
    assert tree2.contains_var('tau')
    bc('emiss')  # just checking it doesn't crash
    with pytest.raises(pc.DimensionalityError):  # can't get emiss_trac for 2D box! needs 3D.
        bc('emiss', emiss_mode='trac_noopa')
    with pytest.raises(pc.FluidValueError):  # don't do emiss for multifluid!
        bm('emiss')
    bm('SF_emiss')  # multifluid calculator can specify single-fluid mode to get emiss.


def test_bifrost_vdem_vars():
    '''testing vdem-related vars and options, using Bifrost data'''
    bc = get_bifrost_calculator()
    # ensure can get vdems without crash.
    # (this is a tiny simulation box otherwise these would be wayyyy too expensive to do quickly):
    assert len(bc.VDEM_MODE_OPTIONS) >= 1  # i.e., the loop below is nontrivial...
    for x in ['x', 'z']:
        for vdem_mode in bc.VDEM_MODE_OPTIONS:
            bc(f'vdem_{x}', vdem_mode=vdem_mode)  # just checking it doesn't crash.
    # ensure vdem does crash if multiple components
    with pytest.raises(pc.ComponentValueError):
        bc('vdem', component=['x', 'z'])
    # check some dependencies
    bc.component = 'z'  # let's just stick with vdem_z for everything below.
    bc.tree('vdem')  # just checking it doesn't crash
    assert bc.tree('vdem_no_interp').contains_var('vdem')   # vdem_no_interp dispatches to vdem
    assert bc.tree('vdem_interp').contains_var('vdem')      # similar for interp
    assert bc.tree('vdem_allinterp').contains_var('vdem')   # similar for allinterp
    # vdem-related options
    assert np.all(bc('rcoords_logT') == bc.rcoords_logT)
    assert np.all(bc('rcoords_vdop', units='si') == bc.rcoords_vdop_kms * 1e3)
    # ensure can't set vdem options to invalid values:
    bc.rcoords_logT = np.arange(4.5, 6, 0.1)
    bc.vdem_logT_min = 4
    with pytest.raises(pc.InputError):
        bc.rcoords_logT = np.arange(3, 5, 0.1)  # can't set logT with minimum < vdem_logT_min
    with pytest.raises(pc.InputError):
        bc.vdem_logT_min = 5  # can't set vdem_logT_min > min(logT)
    with bc.using(vdem_logT=(5, [6.0,6.4,6.8,7.2])):
        pass  # just want to ensure that's valid and doesn't crash^^
        # and that it restores the previous values afterwards:
    assert np.all(bc.rcoords_logT == np.arange(4.5, 6, 0.1))
    assert np.all(bc.vdem_logT_min == 4)
    # set appropriate options for the tinybox region selected here,
    #  then double-check that they were actually updated
    bc.set_attrs(component='z', vdem_mode='interp', vdem_logT_min=0,
                 rcoords_logT=np.arange(3, 5, 0.2),
                 rcoords_vdop_kms=np.arange(-20, 20, 5))
    assert np.all(bc('rcoords_logT') == np.arange(3, 5, 0.2))
    assert np.all(bc('rcoords_vdop', units='si') == bc.rcoords_vdop_kms * 1e3)
    # compute vdem and double-check some things
    vdem0 = bc('vdem')
    assert np.all(vdem0.sel(logT=slice(None, 3.3)) == 0)  # nothing in tinybox at "very cold" temps
    assert np.all(vdem0.sel(logT=slice(4.3, None)) == 0)  # nothing in tinybox at "very hot" temps
    assert not np.all(vdem0.sel(logT=slice(3.3, 4.3)) == 0)  # something nonzero in the middle.
    assert np.all(vdem0.sel(vdop=slice(None, -12)) == 0)  # nothing in tinybox at "very fast" -u
    assert np.all(vdem0.sel(vdop=slice(12, None)) == 0)  # nothing in tinybox at "very fast" +u
    assert not np.all(vdem0.sel(vdop=slice(-12, 12)) == 0)  # something nonzero in the middle.
    # hard-coded: computed some values. Hopefully they don't change.
    # (if they do change, test will fail. Then can choose if test should be updated, or if it was a bug.)
    # (giving a little bit of buffer room in case of different floating point rounding schemes.)
    # Changelog...:
    #    (2025-09-30) current values, see below!
    #    (2025-09-22) values were smaller, by a factor of ~7000.
    #       Related to a bug (fixed in commit b786d6c) in dx,dy,dz mesh values for Bifrost.
    vstats = {  # compute all as dict first <--> if test fails here, crash will be easier to understand.
        'max': vdem0.max().item(),
        'min': vdem0.where(vdem0 > 0).min().item(),
        'mean': vdem0.mean().item(),
        'median': vdem0.where(vdem0 > 0).median().item(),
        'count': vdem0.where(vdem0 > 0).count().item(),  # number of nonzero points in vdem
    }
    assert 3.2e4 < vstats['max'] < 3.3e4
    assert 1.8e-1 < vstats['min'] < 1.9e-1
    assert 2.0e2 < vstats['mean'] < 2.1e2
    assert 3.1e2 < vstats['median'] < 3.2e2
    assert vstats['count'] == 325


def test_bifrost_Gofnt_vars():
    '''testing G(T)-related vars and options, using Bifrost data'''
    bc = get_bifrost_calculator()
    bc.tree('GofT_fe_9_ion')  # just checking it doesn't crash
    tree = bc.tree('spectra_fe_9_ion_prof')
    assert tree.contains_var('GofT_fe_9_ion_em')
    # [TODO] more tests with GofT.
    # maybe including a test which actually computes values?
    #  The problem is, it depends on muse & chiantipy, which are not PlasmaCalcs dependencies.
    #  So, might not be able to do it in PlasmaCalcs autotests.
